@@ -28,11 +28,9 @@
 #include <Stream.h>
 #include <FontInfo.h>
 #include <PDFDocEncoding.h>
-#include <UGooString.h>
 
 #include "poppler.h"
 #include "poppler-private.h"
-#include "poppler-attachment.h"
 #include "poppler-enums.h"
 
 enum {
@@ -51,6 +49,7 @@ enum {
 	PROP_PAGE_MODE,
 	PROP_VIEWER_PREFERENCES,
 	PROP_PERMISSIONS,
+	PROP_METADATA
 };
 
 typedef struct _PopplerDocumentClass PopplerDocumentClass;
@@ -61,50 +60,14 @@ struct _PopplerDocumentClass
 
 G_DEFINE_TYPE (PopplerDocument, poppler_document, G_TYPE_OBJECT);
 
-/**
- * poppler_document_new_from_file:
- * @uri: uri of the file to load
- * @password: password to unlock the file with, or %NULL
- * @error: Return location for an error, or %NULL
- * 
- * Creates a new #PopplerDocument.  If %NULL is returned, then @error will be
- * set. Possible errors include those in the #POPPLER_ERROR and #G_FILE_ERROR
- * domains.
- * 
- * Return value: A newly created #PopplerDocument, or %NULL
- **/
-PopplerDocument *
-poppler_document_new_from_file (const char  *uri,
-				const char  *password,
-				GError     **error)
+static PopplerDocument *
+_poppler_document_new_from_pdfdoc (PDFDoc  *newDoc,
+                                   GError **error)
 {
   PopplerDocument *document;
-  PDFDoc *newDoc;
-  GooString *filename_g;
-  GooString *password_g;
   int err;
-  char *filename;
 
   document = (PopplerDocument *) g_object_new (POPPLER_TYPE_DOCUMENT, NULL, NULL);
-  
-  if (!globalParams) {
-    globalParams = new GlobalParams("/etc/xpdfrc");
-  }
-
-  filename = g_filename_from_uri (uri, NULL, error);
-  if (!filename)
-    return NULL;
-
-  filename_g = new GooString (filename);
-  g_free (filename);
-
-  password_g = NULL;
-  if (password != NULL)
-    password_g = new GooString (password);
-
-  newDoc = new PDFDoc(filename_g, password_g, password_g);
-  if (password_g)
-    delete password_g;
 
   if (!newDoc->isOk()) {
     err = newDoc->getErrorCode();
@@ -116,9 +79,8 @@ poppler_document_new_from_file (const char  *uri,
     } else {
       g_set_error (error, G_FILE_ERROR,
 		   G_FILE_ERROR_FAILED,
-		   "Failed to load document (error %d) '%s'\n",
-		   err,
-		   uri);
+		   "Failed to load document from data (error %d)'\n",
+		   err);
     }
 
     return NULL;
@@ -140,6 +102,96 @@ poppler_document_new_from_file (const char  *uri,
   return document;
 }
 
+/**
+ * poppler_document_new_from_file:
+ * @uri: uri of the file to load
+ * @password: password to unlock the file with, or %NULL
+ * @error: Return location for an error, or %NULL
+ * 
+ * Creates a new #PopplerDocument.  If %NULL is returned, then @error will be
+ * set. Possible errors include those in the #POPPLER_ERROR and #G_FILE_ERROR
+ * domains.
+ * 
+ * Return value: A newly created #PopplerDocument, or %NULL
+ **/
+PopplerDocument *
+poppler_document_new_from_file (const char  *uri,
+				const char  *password,
+				GError     **error)
+{
+  PDFDoc *newDoc;
+  GooString *filename_g;
+  GooString *password_g;
+  int err;
+  char *filename;
+
+  if (!globalParams) {
+    globalParams = new GlobalParams();
+  }
+
+  filename = g_filename_from_uri (uri, NULL, error);
+  if (!filename)
+    return NULL;
+
+  filename_g = new GooString (filename);
+  g_free (filename);
+
+  password_g = NULL;
+  if (password != NULL)
+    password_g = new GooString (password);
+
+  newDoc = new PDFDoc(filename_g, password_g, password_g);
+  if (password_g)
+    delete password_g;
+
+  return _poppler_document_new_from_pdfdoc (newDoc, error);
+}
+
+/**
+ * poppler_document_new_from_data:
+ * @data: the pdf data contained in a char array
+ * @length: the length of #data
+ * @password: password to unlock the file with, or %NULL
+ * @error: Return location for an error, or %NULL
+ * 
+ * Creates a new #PopplerDocument.  If %NULL is returned, then @error will be
+ * set. Possible errors include those in the #POPPLER_ERROR and #G_FILE_ERROR
+ * domains.
+ * 
+ * Return value: A newly created #PopplerDocument, or %NULL
+ **/
+PopplerDocument *
+poppler_document_new_from_data (char        *data,
+                                int          length,
+                                const char  *password,
+                                GError     **error)
+{
+  Object obj;
+  PDFDoc *newDoc;
+  MemStream *str;
+  GooString *password_g;
+  int err;
+  char *filename;
+
+  if (!globalParams) {
+    globalParams = new GlobalParams();
+  }
+  
+  // create stream
+  obj.initNull();
+  str = new MemStream(data, 0, length, &obj);
+
+  password_g = NULL;
+  if (password != NULL)
+    password_g = new GooString (password);
+
+  newDoc = new PDFDoc(str, password_g, password_g);
+  if (password_g)
+    delete password_g;
+
+  return _poppler_document_new_from_pdfdoc (newDoc, error);
+}
+
 /**
  * poppler_document_save:
  * @document: a #PopplerDocument
@@ -164,6 +216,7 @@ poppler_document_save (PopplerDocument  *document,
   filename = g_filename_from_uri (uri, NULL, error);
   if (filename != NULL) {
     GooString *fname = new GooString (filename);
+    g_free (filename);
 
     retval = document->doc->saveAs (fname);
 
@@ -311,6 +364,7 @@ poppler_document_get_attachments (PopplerDocument *document)
 
       emb_file = catalog->embeddedFile (i);
       attachment = _poppler_attachment_new (document, emb_file);
+      delete emb_file;
 
       retval = g_list_prepend (retval, attachment);
     }
@@ -334,12 +388,12 @@ poppler_document_find_dest (PopplerDocument *document,
 {
 	PopplerDest *dest = NULL;
 	LinkDest *link_dest = NULL;
-	UGooString *g_link_name;
+	GooString *g_link_name;
 
 	g_return_val_if_fail (POPPLER_IS_DOCUMENT (document), NULL);
 	g_return_val_if_fail (link_name != NULL, NULL);
 
-	g_link_name = new UGooString (link_name);
+	g_link_name = new GooString (link_name);
 
 	if (g_link_name) {
 		link_dest = document->doc->findDest (g_link_name);
@@ -409,11 +463,6 @@ static void
 info_dict_get_date (Dict *info_dict, const gchar *key, GValue *value) 
 {
   Object obj;
-  GooString *goo_value;
-  int year, mon, day, hour, min, sec;
-  int scanned_items;
-  struct tm *time;
-  gchar *date_string, *ds;
   GTime result;
 
   if (!info_dict->lookup ((gchar *)key, &obj)->isString ()) {
@@ -421,73 +470,10 @@ info_dict_get_date (Dict *info_dict, const gchar *key, GValue *value)
     return;
   }
 
-  goo_value = obj.getString (); 
+  if (_poppler_convert_pdf_date_to_gtime (obj.getString (), &result))
+    g_value_set_int (value, result);
 
-  if (goo_value->hasUnicodeMarker()) {
-    date_string = g_convert (goo_value->getCString () + 2,
-			goo_value->getLength () - 2,
-			"UTF-8", "UTF-16BE", NULL, NULL, NULL);		
-  } else {
-    date_string = g_strndup (goo_value->getCString (), goo_value->getLength ());
-  }
-  ds = date_string;
-  
-  /* See PDF Reference 1.3, Section 3.8.2 for PDF Date representation */
-
-  if (date_string [0] == 'D' && date_string [1] == ':')
-		date_string += 2;
-	
-  /* FIXME only year is mandatory; parse optional timezone offset */
-  scanned_items = sscanf (date_string, "%4d%2d%2d%2d%2d%2d",
-		&year, &mon, &day, &hour, &min, &sec);
-
-  if (scanned_items != 6) {
-    obj.free ();
-    g_free (ds);
-    return;
-  }
-  
-  /* Workaround for y2k bug in Distiller 3, hoping that it won't
-   * be used after y2.2k */
-  if (year < 1930 && strlen (date_string) > 14) {
-    int century, years_since_1900;
-    scanned_items = sscanf (date_string, "%2d%3d%2d%2d%2d%2d%2d",
-		&century, &years_since_1900, &mon, &day, &hour, &min, &sec);
-						
-    if (scanned_items != 7) {
-      obj.free ();
-      g_free (ds);
-      return;
-    }
-    
-    year = century * 100 + years_since_1900;
-  }
-
-  time = g_new0 (struct tm, 1);
-	
-  time->tm_year = year - 1900;
-  time->tm_mon = mon - 1;
-  time->tm_mday = day;
-  time->tm_hour = hour;
-  time->tm_min = min;
-  time->tm_sec = sec;
-  time->tm_wday = -1;
-  time->tm_yday = -1;
-  time->tm_isdst = -1; /* 0 = DST off, 1 = DST on, -1 = don't know */
- 
-  /* compute tm_wday and tm_yday and check date */
-  if (mktime (time) == (time_t) - 1) {
-    obj.free ();
-    g_free (ds);
-    return;
-  } else {
-  	result = mktime (time);
-  }       
-    
   obj.free ();
-  g_free (ds);
-  
-  g_value_set_int (value, result);
 }
 
 static PopplerPageLayout
@@ -641,6 +627,17 @@ poppler_document_get_property (GObject    *object,
 	flag |= POPPLER_PERMISSIONS_OK_TO_ADD_NOTES;
       g_value_set_flags (value, flag);
       break;
+    case PROP_METADATA:
+      catalog = document->doc->getCatalog ();
+      if (catalog && catalog->isOk ())
+	{
+	  GooString *s = catalog->readMetadata ();
+	  if ( s != NULL ) {
+	  	g_value_set_string (value, s->getCString());
+	  	delete s;
+	  }
+	}
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
     }
@@ -783,6 +780,15 @@ poppler_document_class_init (PopplerDocumentClass *klass)
 			       POPPLER_TYPE_PERMISSIONS,
 			       POPPLER_PERMISSIONS_FULL,
 			       G_PARAM_READABLE));
+
+  g_object_class_install_property
+	  (G_OBJECT_CLASS (klass),
+	   PROP_METADATA,
+	   g_param_spec_string ("metadata",
+				"XML Metadata",
+				"Embedded XML metadata",
+				NULL,
+				G_PARAM_READABLE));
 }
 
 static void
@@ -1096,6 +1102,22 @@ poppler_fonts_iter_get_name (PopplerFontsIter *iter)
 	return name;
 }
 
+const char *
+poppler_fonts_iter_get_file_name (PopplerFontsIter *iter)
+{
+	GooString *file;
+	FontInfo *info;
+
+	info = (FontInfo *)iter->items->get (iter->index);
+
+	file = info->getFile();
+	if (file != NULL) {
+		return file->getCString();
+	} else {
+		return NULL;
+	}
+}
+
 PopplerFontType
 poppler_fonts_iter_get_font_type (PopplerFontsIter *iter)
 {
@@ -1182,6 +1204,42 @@ poppler_fonts_iter_new (GooList *items)
 	return iter;
 }
 
+
+typedef struct _PopplerFontInfoClass PopplerFontInfoClass;
+struct _PopplerFontInfoClass
+{
+        GObjectClass parent_class;
+};
+
+G_DEFINE_TYPE (PopplerFontInfo, poppler_font_info, G_TYPE_OBJECT);
+
+static void poppler_font_info_finalize (GObject *object);
+
+
+static void
+poppler_font_info_class_init (PopplerFontInfoClass *klass)
+{
+        GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
+
+        gobject_class->finalize = poppler_font_info_finalize;
+}
+
+static void
+poppler_font_info_init (PopplerFontInfo *font_info)
+{
+        font_info->document = NULL;
+        font_info->scanner = NULL;
+}
+
+static void
+poppler_font_info_finalize (GObject *object)
+{
+        PopplerFontInfo *font_info = POPPLER_FONT_INFO (object);
+
+        delete font_info->scanner;
+        g_object_unref (font_info->document);
+}
+
 PopplerFontInfo *
 poppler_font_info_new (PopplerDocument *document)
 {
@@ -1189,7 +1247,8 @@ poppler_font_info_new (PopplerDocument *document)
 
 	g_return_val_if_fail (POPPLER_IS_DOCUMENT (document), NULL);
 
-	font_info = g_new0 (PopplerFontInfo, 1);
+	font_info = (PopplerFontInfo *) g_object_new (POPPLER_TYPE_FONT_INFO,
+						      NULL);
 	font_info->document = (PopplerDocument *) g_object_ref (document);
 	font_info->scanner = new FontInfoScanner(document->doc);
 
@@ -1219,14 +1278,13 @@ poppler_font_info_scan (PopplerFontInfo   *font_info,
 	return (items != NULL);
 }
 
+/* For backward compatibility */
 void
 poppler_font_info_free (PopplerFontInfo *font_info)
 {
 	g_return_if_fail (font_info != NULL);
 
-	delete font_info->scanner;
-	g_object_unref (font_info->document);
-	g_free (font_info);
+	g_object_unref (font_info);
 }
 
 
@@ -1348,3 +1406,107 @@ poppler_ps_file_free (PopplerPSFile *ps_file)
 	g_return_if_fail (ps_file != NULL);
         g_object_unref (ps_file);
 }
+
+/**
+ * poppler_document_get_form_field:
+ * @document: a #PopplerDocument
+ * @id: an id of a #PopplerFormField
+ *
+ * Returns the #PopplerFormField for the given @id. It must be freed with
+ * g_object_unref()
+ *
+ * Return value: a new #PopplerFormField or NULL if not found
+ **/
+PopplerFormField *
+poppler_document_get_form_field (PopplerDocument *document,
+				 gint             id)
+{
+  Catalog *catalog = document->doc->getCatalog();
+  unsigned pageNum;
+  unsigned fieldNum;
+  FormPageWidgets *widgets;
+  FormWidget *field;
+
+  FormWidget::decodeID (id, &pageNum, &fieldNum);
+  
+  widgets = catalog->getPage (pageNum)->getPageWidgets ();
+  if (!widgets)
+    return NULL;
+  
+  field = widgets->getWidget (fieldNum);
+  if (field)
+    return _poppler_form_field_new (document, field);
+
+  return NULL;
+}
+
+gboolean
+_poppler_convert_pdf_date_to_gtime (GooString *date,
+				    GTime     *gdate) 
+{
+  int year, mon, day, hour, min, sec;
+  int scanned_items;
+  struct tm time;
+  gchar *date_string, *ds;
+  GTime result;
+
+  if (date->hasUnicodeMarker()) {
+    date_string = g_convert (date->getCString () + 2,
+			     date->getLength () - 2,
+			     "UTF-8", "UTF-16BE", NULL, NULL, NULL);		
+  } else {
+    date_string = g_strndup (date->getCString (), date->getLength ());
+  }
+  ds = date_string;
+  
+  /* See PDF Reference 1.3, Section 3.8.2 for PDF Date representation */
+
+  if (date_string [0] == 'D' && date_string [1] == ':')
+    date_string += 2;
+	
+  /* FIXME only year is mandatory; parse optional timezone offset */
+  scanned_items = sscanf (date_string, "%4d%2d%2d%2d%2d%2d",
+			  &year, &mon, &day, &hour, &min, &sec);
+  
+  if (scanned_items != 6) {
+    g_free (ds);
+    return FALSE;
+  }
+  
+  /* Workaround for y2k bug in Distiller 3, hoping that it won't
+   * be used after y2.2k */
+  if (year < 1930 && strlen (date_string) > 14) {
+    int century, years_since_1900;
+    scanned_items = sscanf (date_string, "%2d%3d%2d%2d%2d%2d%2d",
+			    &century, &years_since_1900, &mon, &day, &hour, &min, &sec);
+						
+    if (scanned_items != 7) {
+      g_free (ds);
+      return FALSE;
+    }
+    
+    year = century * 100 + years_since_1900;
+  }
+
+  time.tm_year = year - 1900;
+  time.tm_mon = mon - 1;
+  time.tm_mday = day;
+  time.tm_hour = hour;
+  time.tm_min = min;
+  time.tm_sec = sec;
+  time.tm_wday = -1;
+  time.tm_yday = -1;
+  time.tm_isdst = -1; /* 0 = DST off, 1 = DST on, -1 = don't know */
+ 
+  /* compute tm_wday and tm_yday and check date */
+  result = mktime (&time);
+  if (result == (time_t) - 1) {
+    g_free (ds);
+    return FALSE;
+  }
+    
+  g_free (ds);
+  *gdate = result;
+
+  return TRUE;
+}
