@@ -38,7 +38,7 @@ typedef enum
   POPPLER_PAGE_LAYOUT_TWO_COLUMN_LEFT,
   POPPLER_PAGE_LAYOUT_TWO_COLUMN_RIGHT,
   POPPLER_PAGE_LAYOUT_TWO_PAGE_LEFT,
-  POPPLER_PAGE_LAYOUT_TWO_PAGE_RIGHT,
+  POPPLER_PAGE_LAYOUT_TWO_PAGE_RIGHT
 } PopplerPageLayout;
 
 typedef enum
@@ -49,7 +49,7 @@ typedef enum
   POPPLER_PAGE_MODE_USE_THUMBS,
   POPPLER_PAGE_MODE_FULL_SCREEN,
   POPPLER_PAGE_MODE_USE_OC,
-  POPPLER_PAGE_MODE_USE_ATTACHMENTS,
+  POPPLER_PAGE_MODE_USE_ATTACHMENTS
 } PopplerPageMode;
 
 typedef enum
@@ -57,11 +57,15 @@ typedef enum
   POPPLER_FONT_TYPE_UNKNOWN,
   POPPLER_FONT_TYPE_TYPE1,
   POPPLER_FONT_TYPE_TYPE1C,
+  POPPLER_FONT_TYPE_TYPE1COT,
   POPPLER_FONT_TYPE_TYPE3,
   POPPLER_FONT_TYPE_TRUETYPE,
+  POPPLER_FONT_TYPE_TRUETYPEOT,
   POPPLER_FONT_TYPE_CID_TYPE0,
   POPPLER_FONT_TYPE_CID_TYPE0C,
-  POPPLER_FONT_TYPE_CID_TYPE2
+  POPPLER_FONT_TYPE_CID_TYPE0COT,
+  POPPLER_FONT_TYPE_CID_TYPE2,
+  POPPLER_FONT_TYPE_CID_TYPE2OT
 } PopplerFontType;
 
 typedef enum /*< flags >*/
@@ -73,7 +77,7 @@ typedef enum /*< flags >*/
   POPPLER_VIEWER_PREFERENCES_FIT_WINDOW = 1 << 3,
   POPPLER_VIEWER_PREFERENCES_CENTER_WINDOW = 1 << 4,
   POPPLER_VIEWER_PREFERENCES_DISPLAY_DOC_TITLE = 1 << 5,
-  POPPLER_VIEWER_PREFERENCES_DIRECTION_RTL = 1 << 6,
+  POPPLER_VIEWER_PREFERENCES_DIRECTION_RTL = 1 << 6
 } PopplerViewerPreferences;
 
 typedef enum /*< flags >*/
@@ -82,7 +86,7 @@ typedef enum /*< flags >*/
   POPPLER_PERMISSIONS_OK_TO_MODIFY = 1 << 1,
   POPPLER_PERMISSIONS_OK_TO_COPY = 1 << 2,
   POPPLER_PERMISSIONS_OK_TO_ADD_NOTES = 1 << 3,
-  POPPLER_PERMISSIONS_FULL = (POPPLER_PERMISSIONS_OK_TO_PRINT | POPPLER_PERMISSIONS_OK_TO_MODIFY | POPPLER_PERMISSIONS_OK_TO_COPY | POPPLER_PERMISSIONS_OK_TO_ADD_NOTES),
+  POPPLER_PERMISSIONS_FULL = (POPPLER_PERMISSIONS_OK_TO_PRINT | POPPLER_PERMISSIONS_OK_TO_MODIFY | POPPLER_PERMISSIONS_OK_TO_COPY | POPPLER_PERMISSIONS_OK_TO_ADD_NOTES)
 
 } PopplerPermissions;
 
@@ -92,6 +96,10 @@ GType            poppler_document_get_type          (void) G_GNUC_CONST;
 PopplerDocument *poppler_document_new_from_file     (const char       *uri,
 						     const char       *password,
 						     GError          **error);
+PopplerDocument *poppler_document_new_from_data     (char             *data,
+						     int               length,
+						     const char       *password,
+						     GError          **error);
 gboolean         poppler_document_save              (PopplerDocument  *document,
 						     const char       *uri,
 						     GError          **error);
@@ -104,9 +112,16 @@ PopplerPage     *poppler_document_get_page_by_label (PopplerDocument  *document,
 gboolean         poppler_document_has_attachments   (PopplerDocument  *document);
 GList           *poppler_document_get_attachments   (PopplerDocument  *document);
 
+/* Links */
+PopplerDest     *poppler_document_find_dest         (PopplerDocument  *document,
+						     const gchar      *link_name);
 
+/* Form */
+PopplerFormField *poppler_document_get_form_field   (PopplerDocument  *document,
+						     gint              id);
 
 /* Interface for getting the Index of a poppler_document */
+#define POPPLER_TYPE_INDEX_ITER                 (poppler_index_iter_get_type ())
 GType             poppler_index_iter_get_type   (void) G_GNUC_CONST;
 PopplerIndexIter *poppler_index_iter_new        (PopplerDocument   *document);
 PopplerIndexIter *poppler_index_iter_copy       (PopplerIndexIter  *iter);
@@ -118,23 +133,33 @@ PopplerAction    *poppler_index_iter_get_action (PopplerIndexIter  *iter);
 gboolean          poppler_index_iter_next       (PopplerIndexIter  *iter);
 
 /* Interface for getting the Fonts of a poppler_document */
+#define POPPLER_TYPE_FONT_INFO             (poppler_font_info_get_type ())
+#define POPPLER_FONT_INFO(obj)             (G_TYPE_CHECK_INSTANCE_CAST ((obj), POPPLER_TYPE_FONT_INFO, PopplerFontInfo))
+#define POPPLER_IS_FONT_INFO(obj)          (G_TYPE_CHECK_INSTANCE_TYPE ((obj), POPPLER_TYPE_FONT_INFO))
+GType             poppler_font_info_get_type       (void) G_GNUC_CONST;
 PopplerFontInfo  *poppler_font_info_new            (PopplerDocument   *document);
 gboolean          poppler_font_info_scan           (PopplerFontInfo   *font_info,
 						    int                n_pages,
 						    PopplerFontsIter **iter);
-void		  poppler_font_info_free	   (PopplerFontInfo   *font_info);
+void             poppler_font_info_free            (PopplerFontInfo   *font_info);
 
+#define POPPLER_TYPE_FONTS_ITER                    (poppler_fonts_iter_get_type ())
 GType             poppler_fonts_iter_get_type      (void) G_GNUC_CONST;
 PopplerFontsIter *poppler_fonts_iter_copy          (PopplerFontsIter  *iter);
 void              poppler_fonts_iter_free          (PopplerFontsIter  *iter);
 const char       *poppler_fonts_iter_get_name      (PopplerFontsIter  *iter);
 const char       *poppler_fonts_iter_get_full_name (PopplerFontsIter  *iter);
+const char       *poppler_fonts_iter_get_file_name (PopplerFontsIter  *iter);
 PopplerFontType   poppler_fonts_iter_get_font_type (PopplerFontsIter  *iter);
 gboolean	  poppler_fonts_iter_is_embedded   (PopplerFontsIter  *iter);
 gboolean	  poppler_fonts_iter_is_subset     (PopplerFontsIter  *iter);
 gboolean          poppler_fonts_iter_next          (PopplerFontsIter  *iter);
 
 /* Export to ps */
+#define POPPLER_TYPE_PS_FILE             (poppler_ps_file_get_type ())
+#define POPPLER_PS_FILE(obj)             (G_TYPE_CHECK_INSTANCE_CAST ((obj), POPPLER_TYPE_PS_FILE, PopplerPSFile))
+#define POPPLER_IS_PS_FILE(obj)          (G_TYPE_CHECK_INSTANCE_TYPE ((obj), POPPLER_TYPE_PS_FILE))
+GType          poppler_ps_file_get_type       (void) G_GNUC_CONST;
 PopplerPSFile *poppler_ps_file_new            (PopplerDocument *document,
                                                const char      *filename,
                                                int              first_page,
