@@ -27,6 +27,7 @@
 #include <ErrorCodes.h>
 #include <UnicodeMap.h>
 #include <GfxState.h>
+#include <PageTransition.h>
 
 #include "poppler.h"
 #include "poppler-private.h"
@@ -72,7 +73,10 @@ poppler_page_finalize (GObject *object)
     delete page->gfx;
   if (page->text_dev != NULL)
     delete page->text_dev;
-
+#if defined (HAVE_CAIRO)
+  if (page->image_dev != NULL)
+    delete page->image_dev;
+#endif
   /* page->page is owned by the document */
 }
 
@@ -125,6 +129,109 @@ poppler_page_get_index (PopplerPage *page)
   return page->index;
 }
 
+/**
+ * poppler_page_get_duration:
+ * @page: a #PopplerPage
+ *
+ * Returns the duration of @page
+ *
+ * Return value: duration in seconds of @page or -1.
+ **/
+double
+poppler_page_get_duration (PopplerPage *page)
+{
+  g_return_val_if_fail (POPPLER_IS_PAGE (page), -1);
+
+  return page->page->getDuration ();
+}
+
+/**
+ * poppler_page_get_transition:
+ * @page: a #PopplerPage
+ *
+ * Returns the transition effect of @page
+ *
+ * Return value: a #PopplerPageTransition or NULL.
+ **/
+PopplerPageTransition *
+poppler_page_get_transition (PopplerPage *page)
+{
+  PageTransition *trans;
+  PopplerPageTransition *transition;
+  Object obj;
+  
+  g_return_val_if_fail (POPPLER_IS_PAGE (page), NULL);
+
+  trans = new PageTransition (page->page->getTrans (&obj));
+  obj.free ();
+
+  if (!trans->isOk ()) {
+    delete trans;
+    return NULL;
+  }
+
+  transition = poppler_page_transition_new ();
+
+  switch (trans->getType ())
+    {
+      case transitionReplace:
+        transition->type = POPPLER_PAGE_TRANSITION_REPLACE;
+	break;
+      case transitionSplit:
+	transition->type = POPPLER_PAGE_TRANSITION_SPLIT;
+	break;
+      case transitionBlinds:
+        transition->type = POPPLER_PAGE_TRANSITION_BLINDS;
+	break;
+      case transitionBox:
+        transition->type = POPPLER_PAGE_TRANSITION_BOX;
+	break;
+      case transitionWipe:
+        transition->type = POPPLER_PAGE_TRANSITION_WIPE;
+	break;
+      case transitionDissolve:
+        transition->type = POPPLER_PAGE_TRANSITION_DISSOLVE;
+	break;
+      case transitionGlitter:
+        transition->type = POPPLER_PAGE_TRANSITION_GLITTER;
+	break;
+      case transitionFly:
+        transition->type = POPPLER_PAGE_TRANSITION_FLY;
+	break;
+      case transitionPush:
+        transition->type = POPPLER_PAGE_TRANSITION_PUSH;
+	break;
+      case transitionCover:
+        transition->type = POPPLER_PAGE_TRANSITION_COVER;
+	break;
+      case transitionUncover:
+        transition->type = POPPLER_PAGE_TRANSITION_UNCOVER;
+        break;
+      case transitionFade:
+        transition->type = POPPLER_PAGE_TRANSITION_FADE;
+	break;
+      default:
+        g_assert_not_reached ();
+    }
+
+  transition->alignment = (trans->getAlignment() == transitionHorizontal) ?
+	  POPPLER_PAGE_TRANSITION_HORIZONTAL :
+	  POPPLER_PAGE_TRANSITION_VERTICAL;
+
+  transition->direction = (trans->getDirection() == transitionInward) ?
+	  POPPLER_PAGE_TRANSITION_INWARD :
+	  POPPLER_PAGE_TRANSITION_OUTWARD;
+
+  transition->duration = trans->getDuration();
+  transition->angle = trans->getAngle();
+  transition->scale = trans->getScale();
+  transition->rectangular = trans->isRectangular();
+  
+  delete trans;
+  
+  return transition;
+}
+
 #if defined (HAVE_CAIRO)
 
 typedef struct {
@@ -178,9 +285,9 @@ poppler_page_prepare_output_dev (PopplerPage *page,
 }
 
 static void
-poppler_page_copy_to_pixbuf (PopplerPage *page,
-			     GdkPixbuf *pixbuf,
-			     OutputDevData *output_dev_data)
+copy_cairo_surface_to_pixbuf (cairo_surface_t *surface,
+			      unsigned char   *data,
+			      GdkPixbuf       *pixbuf)
 {
   int cairo_width, cairo_height, cairo_rowstride;
   unsigned char *pixbuf_data, *dst, *cairo_data;
@@ -188,10 +295,10 @@ poppler_page_copy_to_pixbuf (PopplerPage *page,
   unsigned int *src;
   int x, y;
 
-  cairo_width = cairo_image_surface_get_width (output_dev_data->surface);
-  cairo_height = cairo_image_surface_get_height (output_dev_data->surface);
+  cairo_width = cairo_image_surface_get_width (surface);
+  cairo_height = cairo_image_surface_get_height (surface);
   cairo_rowstride = cairo_width * 4;
-  cairo_data = output_dev_data->cairo_data;
+  cairo_data = data;
 
   pixbuf_data = gdk_pixbuf_get_pixels (pixbuf);
   pixbuf_rowstride = gdk_pixbuf_get_rowstride (pixbuf);
@@ -216,7 +323,17 @@ poppler_page_copy_to_pixbuf (PopplerPage *page,
 	  src++;
 	}
     }
+}	
 
+static void
+poppler_page_copy_to_pixbuf (PopplerPage *page,
+			     GdkPixbuf *pixbuf,
+			     OutputDevData *output_dev_data)
+{
+  copy_cairo_surface_to_pixbuf (output_dev_data->surface,
+				output_dev_data->cairo_data,
+				pixbuf);
+  
   page->document->output_dev->setCairo (NULL);
   cairo_surface_destroy (output_dev_data->surface);
   cairo_destroy (output_dev_data->cairo);
@@ -309,14 +426,15 @@ poppler_page_render (PopplerPage *page,
   output_dev = page->document->output_dev;
   output_dev->setCairo (cairo);
 
+  /* NOTE: instead of passing -1 we should/could use cairo_clip_extents()
+   * to get a bounding box */
   page->page->displaySlice(output_dev,
 			   72.0, 72.0, 0,
 			   gFalse, /* useMediaBox */
 			   gTrue, /* Crop */
-			   0, 0,
-			   (int) ceil (page->page->getCropWidth ()),
-			   (int) ceil (page->page->getCropHeight ()),
-			   NULL, /* links */
+			   -1, -1,
+			   -1, -1,
+			   gFalse, /* printing */
 			   page->document->doc->getCatalog ());
 
   output_dev->setCairo (NULL);
@@ -325,7 +443,7 @@ poppler_page_render (PopplerPage *page,
 #endif
 
 /**
- * poppler_page_render:
+ * poppler_page_render_to_pixbuf:
  * @page: the page to render from
  * @src_x: x coordinate of upper left corner  
  * @src_y: y coordinate of upper left corner  
@@ -362,9 +480,9 @@ poppler_page_render_to_pixbuf (PopplerPage *page,
 			   gTrue, /* Crop */
 			   src_x, src_y,
 			   src_width, src_height,
-			   NULL, /* links */
+			   gFalse, /* printing */
 			   page->document->doc->getCatalog ());
-
+  
   poppler_page_copy_to_pixbuf (page, pixbuf, &data);
 }
 
@@ -374,12 +492,14 @@ poppler_page_get_text_output_dev (PopplerPage *page)
   if (page->text_dev == NULL) {
     page->text_dev = new TextOutputDev (NULL, gTrue, gFalse, gFalse);
 
+    if (page->gfx)
+      delete page->gfx;
     page->gfx = page->page->createGfx(page->text_dev,
 				      72.0, 72.0, 0,
 				      gFalse, /* useMediaBox */
 				      gTrue, /* Crop */
 				      -1, -1, -1, -1,
-				      NULL, /* links */
+				      gFalse, /* printing */
 				      page->document->doc->getCatalog (),
 				      NULL, NULL, NULL, NULL);
 
@@ -395,21 +515,25 @@ poppler_page_get_text_output_dev (PopplerPage *page)
  * poppler_page_get_selection_region:
  * @page: a #PopplerPage
  * @scale: scale specified as pixels per point
+ * @style: a #PopplerSelectionStyle
  * @selection: start and end point of selection as a rectangle
  * 
  * Returns a region containing the area that would be rendered by
- * poppler_page_render_selection().  The returned region must be freed with
- * gdk_region_destroy().
+ * poppler_page_render_selection() or
+ * poppler_page_render_selection_to_pixbuf(). The returned
+ * region must be freed with gdk_region_destroy().
  * 
  * Return value: a newly allocated #GdkRegion
  **/
 GdkRegion *
-poppler_page_get_selection_region (PopplerPage      *page,
-				   gdouble           scale,
-				   PopplerRectangle *selection)
+poppler_page_get_selection_region (PopplerPage           *page,
+				   gdouble                scale,
+				   PopplerSelectionStyle  style,
+				   PopplerRectangle      *selection)
 {
   TextOutputDev *text_dev;
   PDFRectangle poppler_selection;
+  SelectionStyle selection_style = selectionStyleGlyph;
   GooList *list;
   GdkRectangle rect;
   GdkRegion *region;
@@ -420,8 +544,22 @@ poppler_page_get_selection_region (PopplerPage      *page,
   poppler_selection.x2 = selection->x2;
   poppler_selection.y2 = selection->y2;
 
+  switch (style)
+    {
+      case POPPLER_SELECTION_GLYPH:
+        selection_style = selectionStyleGlyph;
+	break;
+      case POPPLER_SELECTION_WORD:
+        selection_style = selectionStyleWord;
+	break;
+      case POPPLER_SELECTION_LINE:
+        selection_style = selectionStyleLine;
+	break;
+    }
+	      
   text_dev = poppler_page_get_text_output_dev (page);
-  list = text_dev->getSelectionRegion(&poppler_selection, scale);
+  list = text_dev->getSelectionRegion(&poppler_selection, 
+				      selection_style, scale);
 
   region = gdk_region_new();
 
@@ -503,14 +641,99 @@ poppler_page_set_selection_alpha (PopplerPage      *page,
 
 #endif
 
+#if defined (HAVE_CAIRO)
 /**
  * poppler_page_render_selection:
  * @page: the #PopplerPage for which to render selection
+ * @cairo: cairo context to render to
+ * @selection: start and end point of selection as a rectangle
+ * @old_selection: previous selection
+ * @style: a #PopplerSelectionStyle
+ * @glyph_color: color to use for drawing glyphs
+ * @background_color: color to use for the selection background
+ *
+ * Render the selection specified by @selection for @page to
+ * the given cairo context.  The selection will be rendered, using
+ * @glyph_color for the glyphs and @background_color for the selection
+ * background.
+ *
+ * If non-NULL, @old_selection specifies the selection that is already
+ * rendered to @cairo, in which case this function will (some day)
+ * only render the changed part of the selection.
+ **/
+void
+poppler_page_render_selection (PopplerPage           *page,
+			       cairo_t               *cairo,
+			       PopplerRectangle      *selection,
+			       PopplerRectangle      *old_selection,
+			       PopplerSelectionStyle  style, 
+			       GdkColor              *glyph_color,
+			       GdkColor              *background_color)
+{
+  TextOutputDev *text_dev;
+  CairoOutputDev *output_dev;
+  SelectionStyle selection_style = selectionStyleGlyph;
+  PDFRectangle pdf_selection(selection->x1, selection->y1,
+			     selection->x2, selection->y2);
+
+  GfxColor gfx_background_color = {
+      {
+	  background_color->red,
+	  background_color->green,
+	  background_color->blue
+      }
+  };
+  GfxColor gfx_glyph_color = {
+      {
+	  glyph_color->red,
+	  glyph_color->green,
+	  glyph_color->blue
+      }
+  };
+
+  switch (style)
+    {
+      case POPPLER_SELECTION_GLYPH:
+        selection_style = selectionStyleGlyph;
+	break;
+      case POPPLER_SELECTION_WORD:
+        selection_style = selectionStyleWord;
+	break;
+      case POPPLER_SELECTION_LINE:
+        selection_style = selectionStyleLine;
+	break;
+    }
+
+  text_dev = poppler_page_get_text_output_dev (page);
+  output_dev = page->document->output_dev;
+  output_dev->setCairo (cairo);
+
+  text_dev->drawSelection (output_dev, 1.0, 0,
+			   &pdf_selection, selection_style,
+			   &gfx_glyph_color, &gfx_background_color);
+
+  output_dev->setCairo (NULL);
+
+  /* We'll need a function to destroy page->text_dev and page->gfx
+   * when the application wants to get rid of them.
+   *
+   * Two improvements: 1) make GfxFont refcounted and let TextPage and
+   * friends hold a reference to the GfxFonts they need so we can free
+   * up Gfx early.  2) use a TextPage directly when rendering the page
+   * so we don't have to use TextOutputDev and render a second
+   * time. */
+}
+#endif
+
+/**
+ * poppler_page_render_selection_to_pixbuf:
+ * @page: the #PopplerPage for which to render selection
  * @scale: scale specified as pixels per point
  * @rotation: rotate the document by the specified degree
  * @pixbuf: pixbuf to render to
  * @selection: start and end point of selection as a rectangle
  * @old_selection: previous selection
+ * @style: a #PopplerSelectionStyle
  * @glyph_color: color to use for drawing glyphs
  * @background_color: color to use for the selection background
  * 
@@ -524,18 +747,20 @@ poppler_page_set_selection_alpha (PopplerPage      *page,
  * only render the changed part of the selection.
  **/
 void
-poppler_page_render_selection (PopplerPage      *page,
-			       gdouble           scale,
-			       int		 rotation,
-			       GdkPixbuf        *pixbuf,
-			       PopplerRectangle *selection,
-			       PopplerRectangle *old_selection,
-			       GdkColor         *glyph_color,
-			       GdkColor         *background_color)
+poppler_page_render_selection_to_pixbuf (PopplerPage           *page,
+                                         gdouble                scale,
+                                         int                    rotation,
+                                         GdkPixbuf             *pixbuf,
+                                         PopplerRectangle      *selection,
+                                         PopplerRectangle      *old_selection,
+					 PopplerSelectionStyle  style,
+                                         GdkColor              *glyph_color,
+                                         GdkColor              *background_color)
 {
   TextOutputDev *text_dev;
   OutputDev *output_dev;
   OutputDevData data;
+  SelectionStyle selection_style = selectionStyleGlyph;
   PDFRectangle pdf_selection(selection->x1, selection->y1,
 			     selection->x2, selection->y2);
 
@@ -554,12 +779,26 @@ poppler_page_render_selection (PopplerPage      *page,
       }
   };
 
+  switch (style)
+    {
+      case POPPLER_SELECTION_GLYPH:
+        selection_style = selectionStyleGlyph;
+	break;
+      case POPPLER_SELECTION_WORD:
+        selection_style = selectionStyleWord;
+	break;
+      case POPPLER_SELECTION_LINE:
+        selection_style = selectionStyleLine;
+	break;
+    }
+
   text_dev = poppler_page_get_text_output_dev (page);
   output_dev = page->document->output_dev;
 
   poppler_page_prepare_output_dev (page, scale, rotation, TRUE, &data);
 
-  text_dev->drawSelection (output_dev, scale, rotation, &pdf_selection,
+  text_dev->drawSelection (output_dev, scale, rotation,
+			   &pdf_selection, selection_style,
 			   &gfx_glyph_color, &gfx_background_color);
 
   poppler_page_copy_to_pixbuf (page, pixbuf, &data);
@@ -659,22 +898,25 @@ poppler_page_get_thumbnail_size (PopplerPage *page,
 /**
  * poppler_page_get_text:
  * @page: a #PopplerPage
- * @rect: the rectangle including the text
+ * @style: a #PopplerSelectionStyle
+ * @selection: the #PopplerRectangle including the text
  * 
- * Retrieves the contents of the specified rectangle as text 
+ * Retrieves the contents of the specified @selection as text.
  * 
- * Return value: a pointer to the contents of the rectangle
+ * Return value: a pointer to the contents of the @selection
  *               as a string
  **/
 char *
-poppler_page_get_text (PopplerPage      *page,
-		       PopplerRectangle *selection)
+poppler_page_get_text (PopplerPage          *page,
+		       PopplerSelectionStyle style,
+		       PopplerRectangle     *selection)
 {
   TextOutputDev *text_dev;
   PDFDoc *doc;
   GooString *sel_text;
   double height, y1, y2;
   char *result;
+  SelectionStyle selection_style = selectionStyleGlyph;
   PDFRectangle pdf_selection;
 
   g_return_val_if_fail (POPPLER_IS_PAGE (page), FALSE);
@@ -687,8 +929,21 @@ poppler_page_get_text (PopplerPage      *page,
   pdf_selection.y1 = height - selection->y2;
   pdf_selection.x2 = selection->x2;
   pdf_selection.y2 = height - selection->y1;
-  
-  sel_text = text_dev->getSelectionText (&pdf_selection);
+
+  switch (style)
+    {
+      case POPPLER_SELECTION_GLYPH:
+        selection_style = selectionStyleGlyph;
+	break;
+      case POPPLER_SELECTION_WORD:
+        selection_style = selectionStyleWord;
+	break;
+      case POPPLER_SELECTION_LINE:
+        selection_style = selectionStyleLine;
+	break;
+    }
+
+  sel_text = text_dev->getSelectionText (&pdf_selection, selection_style);
   result = g_strdup (sel_text->getCString ());
   delete sel_text;
 
@@ -728,7 +983,7 @@ poppler_page_find_text (PopplerPage *page,
 
   poppler_page_get_size (page, NULL, &height);
   page->page->display (output_dev, 72, 72, 0, gFalse,
-		       gTrue, NULL, doc->getCatalog());
+		       gTrue, gFalse, doc->getCatalog());
   
   matches = NULL;
   xMin = 0;
@@ -754,6 +1009,135 @@ poppler_page_find_text (PopplerPage *page,
   return g_list_reverse (matches);
 }
 
+#if defined (HAVE_CAIRO)
+
+static CairoImageOutputDev *
+poppler_page_get_image_output_dev (PopplerPage *page)
+{
+  if (page->image_dev == NULL) {
+    page->image_dev = new CairoImageOutputDev ();
+
+    if (page->gfx)
+      delete page->gfx;
+    page->gfx = page->page->createGfx(page->image_dev,
+				      72.0, 72.0, 0,
+				      gFalse, /* useMediaBox */
+				      gTrue, /* Crop */
+				      -1, -1, -1, -1,
+				      gFalse, /* printing */
+				      page->document->doc->getCatalog (),
+				      NULL, NULL, NULL, NULL);
+    page->page->display(page->gfx);
+  }
+
+  return page->image_dev;
+}
+
+static GdkPixbuf *
+poppler_page_image_pixbuf_create (PopplerPage *page,
+				  CairoImage  *image)
+{
+  GdkPixbuf *pixbuf;
+  cairo_surface_t *surface;
+
+  surface = image->getImage ();
+  
+  pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB,
+			   FALSE, 8,
+			   cairo_image_surface_get_width (surface),
+			   cairo_image_surface_get_height (surface));
+  
+  copy_cairo_surface_to_pixbuf (surface,
+				cairo_image_surface_get_data (surface),
+				pixbuf);
+
+  return pixbuf;
+}
+
+/**
+ * poppler_page_get_image_mapping:
+ * @page: A #PopplerPage
+ *
+ * Returns a list of #PopplerImageMapping items that map from a
+ * location on @page to a #GdkPixbuf.  This list must be freed
+ * with poppler_page_free_image_mapping() when done.
+ *
+ * Return value: A #GList of #PopplerImageMapping
+ **/
+GList *
+poppler_page_get_image_mapping (PopplerPage *page)
+{
+  GList *map_list = NULL;
+  CairoImageOutputDev *out;
+  gint i;
+  
+  g_return_val_if_fail (POPPLER_IS_PAGE (page), NULL);
+
+  out = poppler_page_get_image_output_dev (page);
+
+  for (i = 0; i < out->getNumImages (); i++) {
+    PopplerImageMapping *mapping;
+    CairoImage *image;
+
+    image = out->getImage (i);
+
+    /* Create the mapping */
+    mapping = g_new (PopplerImageMapping, 1);
+
+    image->getRect (&(mapping->area.x1), &(mapping->area.y1),
+		    &(mapping->area.x2), &(mapping->area.y2));
+    mapping->image = poppler_page_image_pixbuf_create (page, image);
+    
+    mapping->area.x1 -= page->page->getCropBox()->x1;
+    mapping->area.x2 -= page->page->getCropBox()->x1;
+    mapping->area.y1 -= page->page->getCropBox()->y1;
+    mapping->area.y2 -= page->page->getCropBox()->y1;
+
+    map_list = g_list_prepend (map_list, mapping);
+  }
+
+  return map_list;	
+}
+
+static void
+poppler_images_mapping_free (PopplerImageMapping *mapping)
+{
+  g_object_unref (mapping->image);
+  g_free (mapping);
+}
+
+/**
+ * poppler_page_free_image_mapping:
+ * @list: A list of #PopplerImageMapping<!-- -->s
+ *
+ * Frees a list of #PopplerImageMapping<!-- -->s allocated by
+ * poppler_page_get_image_mapping().
+ **/
+void
+poppler_page_free_image_mapping (GList *list)
+{
+  if (list == NULL)
+    return;
+
+  g_list_foreach (list, (GFunc) (poppler_images_mapping_free), NULL);
+  g_list_free (list);
+}
+
+#else
+
+GList *
+poppler_page_get_image_mapping (PopplerPage *page)
+{
+  return NULL;
+}
+
+void
+poppler_page_free_image_mapping (GList *list)
+{
+}
+
+#endif /* HAVE_CAIRO */
+
 /**
  * poppler_page_render_to_ps:
  * @page: a #PopplerPage
@@ -773,10 +1157,11 @@ poppler_page_render_to_ps (PopplerPage   *page,
     ps_file->out = new PSOutputDev (ps_file->filename,
                                     ps_file->document->doc->getXRef(),
                                     ps_file->document->doc->getCatalog(),
+                                    NULL,
                                     ps_file->first_page, ps_file->last_page,
                                     psModePS, (int)ps_file->paper_width,
                                     (int)ps_file->paper_height, ps_file->duplex,
-                                    0, 0, 0, 0, gFalse);
+                                    0, 0, 0, 0, gFalse, gFalse);
 
 
   ps_file->document->doc->displayPage (ps_file->out, page->index + 1, 72.0, 72.0,
@@ -825,7 +1210,6 @@ poppler_page_init (PopplerPage *page)
 {
 }
 
-
 /**
  * poppler_page_get_link_mapping:
  * @page: A #PopplerPage
@@ -839,52 +1223,87 @@ poppler_page_init (PopplerPage *page)
 GList *
 poppler_page_get_link_mapping (PopplerPage *page)
 {
-	GList *map_list = NULL;
-	gint i;
-	Links *links;
-	Object obj;
-
-	g_return_val_if_fail (POPPLER_IS_PAGE (page), NULL);
-
-	links = new Links (page->page->getAnnots (&obj),
-			   page->document->doc->getCatalog ()->getBaseURI ());
-	obj.free ();
-
-	if (links == NULL)
-		return NULL;
-
-	for (i = 0; i < links->getNumLinks (); i++) {
-		PopplerLinkMapping *mapping;
-		LinkAction *link_action;
-		Link *link;
-
-		link = links->getLink (i);
-		link_action = link->getAction ();
-
-		/* Create the mapping */
-		mapping = g_new (PopplerLinkMapping, 1);
-		mapping->action = _poppler_action_new (page->document, link_action, NULL);
-		link->getRect (&(mapping->area.x1), &(mapping->area.y1),
-			       &(mapping->area.x2), &(mapping->area.y2));
-
-		mapping->area.x1 -= page->page->getCropBox()->x1;
-		mapping->area.x2 -= page->page->getCropBox()->x1;
-		mapping->area.y1 -= page->page->getCropBox()->y1;
-		mapping->area.y2 -= page->page->getCropBox()->y1;
-
-		map_list = g_list_prepend (map_list, mapping);
+  GList *map_list = NULL;
+  gint i;
+  Links *links;
+  Object obj;
+  double width, height;
+  
+  g_return_val_if_fail (POPPLER_IS_PAGE (page), NULL);
+  
+  links = new Links (page->page->getAnnots (&obj),
+		     page->document->doc->getCatalog ()->getBaseURI ());
+  obj.free ();
+  
+  if (links == NULL)
+    return NULL;
+  
+  poppler_page_get_size (page, &width, &height);
+  
+  for (i = 0; i < links->getNumLinks (); i++)
+    {
+      PopplerLinkMapping *mapping;
+      PopplerRectangle rect;
+      LinkAction *link_action;
+      Link *link;
+      
+      link = links->getLink (i);
+      link_action = link->getAction ();
+      
+      /* Create the mapping */
+      mapping = g_new (PopplerLinkMapping, 1);
+      mapping->action = _poppler_action_new (page->document, link_action, NULL);
+      
+      link->getRect (&rect.x1, &rect.y1, &rect.x2, &rect.y2);
+      
+      switch (page->page->getRotate ())
+        {
+        case 90:
+	  mapping->area.x1 = rect.y1;
+	  mapping->area.y1 = height - rect.x2;
+	  mapping->area.x2 = mapping->area.x1 + (rect.y2 - rect.y1);
+	  mapping->area.y2 = mapping->area.y1 + (rect.x2 - rect.x1);
+	  
+	  break;
+	case 180:
+	  mapping->area.x1 = width - rect.x2;
+	  mapping->area.y1 = height - rect.y2;
+	  mapping->area.x2 = mapping->area.x1 + (rect.x2 - rect.x1);
+	  mapping->area.y2 = mapping->area.y1 + (rect.y2 - rect.y1);
+	  
+	  break;
+	case 270:
+	  mapping->area.x1 = width - rect.y2;
+	  mapping->area.y1 = rect.x1;
+	  mapping->area.x2 = mapping->area.x1 + (rect.y2 - rect.y1);
+	  mapping->area.y2 = mapping->area.y1 + (rect.x2 - rect.x1);
+	  
+	  break;
+	default:
+	  mapping->area.x1 = rect.x1;
+	  mapping->area.y1 = rect.y1;
+	  mapping->area.x2 = rect.x2;
+	  mapping->area.y2 = rect.y2;
 	}
-
-	delete links;
-
-	return map_list;
+			
+      mapping->area.x1 -= page->page->getCropBox()->x1;
+      mapping->area.x2 -= page->page->getCropBox()->x1;
+      mapping->area.y1 -= page->page->getCropBox()->y1;
+      mapping->area.y2 -= page->page->getCropBox()->y1;
+      
+      map_list = g_list_prepend (map_list, mapping);
+    }
+  
+  delete links;
+  
+  return map_list;
 }
 
 static void
 poppler_mapping_free (PopplerLinkMapping *mapping)
 {
-	poppler_action_free (mapping->action);
-	g_free (mapping);
+  poppler_action_free (mapping->action);
+  g_free (mapping);
 }
 
 /**
@@ -899,11 +1318,74 @@ poppler_mapping_free (PopplerLinkMapping *mapping)
 void
 poppler_page_free_link_mapping (GList *list)
 {
-	if (list == NULL)
-		return;
+  if (list == NULL)
+    return;
 
-	g_list_foreach (list, (GFunc) (poppler_mapping_free), NULL);
-	g_list_free (list);
+  g_list_foreach (list, (GFunc) (poppler_mapping_free), NULL);
+  g_list_free (list);
+}
+
+/**
+ * poppler_page_get_form_field_mapping:
+ * @page: A #PopplerPage
+ *
+ * Returns a list of #PopplerFormFieldMapping items that map from a
+ * location on @page to a form field.  This list must be freed
+ * with poppler_page_free_form_field_mapping() when done.
+ *
+ * Return value: A #GList of #PopplerFormFieldMapping
+ **/
+GList *
+poppler_page_get_form_field_mapping (PopplerPage *page)
+{
+  GList *map_list = NULL;
+  FormPageWidgets *forms;
+  gint i;
+  
+  g_return_val_if_fail (POPPLER_IS_PAGE (page), NULL);
+
+  forms = page->page->getPageWidgets ();
+  if (forms == NULL)
+    return NULL;
+  
+  for (i = 0; i < forms->getNumWidgets (); i++) {
+    PopplerFormFieldMapping *mapping;
+    FormWidget *field;
+
+    mapping = poppler_form_field_mapping_new ();
+    
+    field = forms->getWidget (i);
+
+    mapping->field = _poppler_form_field_new (page->document, field);
+    field->getRect (&(mapping->area.x1), &(mapping->area.y1),
+		    &(mapping->area.x2), &(mapping->area.y2));
+
+    mapping->area.x1 -= page->page->getCropBox()->x1;
+    mapping->area.x2 -= page->page->getCropBox()->x1;
+    mapping->area.y1 -= page->page->getCropBox()->y1;
+    mapping->area.y2 -= page->page->getCropBox()->y1;
+    
+    map_list = g_list_prepend (map_list, mapping);
+  }
+  
+  return map_list;
+}
+
+/**
+ * poppler_page_free_form_field_mapping:
+ * @list: A list of #PopplerFormFieldMapping<!-- -->s
+ *
+ * Frees a list of #PopplerFormFieldMapping<!-- -->s allocated by
+ * poppler_page_get_form_field_mapping().
+ **/
+void
+poppler_page_free_form_field_mapping (GList *list)
+{
+  if (list == NULL)
+    return;
+
+  g_list_foreach (list, (GFunc) poppler_form_field_mapping_free, NULL);
+  g_list_free (list);
 }
 
 /* PopplerRectangle type */
@@ -924,26 +1406,26 @@ poppler_rectangle_get_type (void)
 PopplerRectangle *
 poppler_rectangle_new (void)
 {
-	return g_new0 (PopplerRectangle, 1);
+  return g_new0 (PopplerRectangle, 1);
 }
 
 PopplerRectangle *
 poppler_rectangle_copy (PopplerRectangle *rectangle)
 {
-	PopplerRectangle *new_rectangle;
+  PopplerRectangle *new_rectangle;
 
-	g_return_val_if_fail (rectangle != NULL, NULL);
+  g_return_val_if_fail (rectangle != NULL, NULL);
+  
+  new_rectangle = g_new0 (PopplerRectangle, 1);
+  *new_rectangle = *rectangle;
 
-	new_rectangle = g_new0 (PopplerRectangle, 1);
-	*new_rectangle = *rectangle;
-
-	return new_rectangle;
+  return new_rectangle;
 }
 
 void
 poppler_rectangle_free (PopplerRectangle *rectangle)
 {
-	g_free (rectangle);
+  g_free (rectangle);
 }
 
 /* PopplerLinkMapping type */
@@ -963,28 +1445,162 @@ poppler_link_mapping_get_type (void)
 PopplerLinkMapping *
 poppler_link_mapping_new (void)
 {
-	return (PopplerLinkMapping *) g_new0 (PopplerLinkMapping, 1);
+  return (PopplerLinkMapping *) g_new0 (PopplerLinkMapping, 1);
 }
 
 PopplerLinkMapping *
 poppler_link_mapping_copy (PopplerLinkMapping *mapping)
 {
-	PopplerLinkMapping *new_mapping;
+  PopplerLinkMapping *new_mapping;
 
-	new_mapping = poppler_link_mapping_new ();
+  new_mapping = poppler_link_mapping_new ();
 	
-	*new_mapping = *mapping;
-	if (new_mapping->action)
-		new_mapping->action = poppler_action_copy (new_mapping->action);
+  *new_mapping = *mapping;
+  if (new_mapping->action)
+    new_mapping->action = poppler_action_copy (new_mapping->action);
 
-	return new_mapping;
+  return new_mapping;
 }
 
 void
 poppler_link_mapping_free (PopplerLinkMapping *mapping)
 {
-	if (mapping)
-		poppler_action_free (mapping->action);
+  if (mapping->action)
+    poppler_action_free (mapping->action);
 
-	g_free (mapping);
+  g_free (mapping);
 }
+
+/* Poppler Image mapping type */
+GType
+poppler_image_mapping_get_type (void)
+{
+  static GType our_type = 0;
+
+  if (our_type == 0)
+    our_type = g_boxed_type_register_static ("PopplerImageMapping",
+					     (GBoxedCopyFunc) poppler_image_mapping_copy,
+					     (GBoxedFreeFunc) poppler_image_mapping_free);
+
+  return our_type;
+}
+
+PopplerImageMapping *
+poppler_image_mapping_new (void)
+{
+  return (PopplerImageMapping *) g_new0 (PopplerImageMapping, 1);
+}
+
+PopplerImageMapping *
+poppler_image_mapping_copy (PopplerImageMapping *mapping)
+{
+  PopplerImageMapping *new_mapping;
+
+  new_mapping = poppler_image_mapping_new ();
+
+  *new_mapping = *mapping;
+  if (new_mapping->image)
+    new_mapping->image = gdk_pixbuf_copy (new_mapping->image);
+
+  return new_mapping;
+}
+
+void
+poppler_image_mapping_free (PopplerImageMapping *mapping)
+{
+  if (mapping->image)
+    g_object_unref (mapping->image);
+
+  g_free (mapping);
+}
+
+/* Page Transition */
+GType
+poppler_page_transition_get_type (void)
+{
+  static GType our_type = 0;
+  if (our_type == 0)
+    our_type = g_boxed_type_register_static("PopplerPageTransition",
+					    (GBoxedCopyFunc) poppler_page_transition_copy,
+					    (GBoxedFreeFunc) poppler_page_transition_free);
+  return our_type;
+}
+
+PopplerPageTransition *
+poppler_page_transition_new (void)
+{
+  return (PopplerPageTransition *) g_new0 (PopplerPageTransition, 1);
+}
+
+PopplerPageTransition *
+poppler_page_transition_copy (PopplerPageTransition *transition)
+{
+  PopplerPageTransition *new_transition;
+
+  new_transition = poppler_page_transition_new ();
+  *new_transition = *transition;
+  
+  return new_transition;
+}
+
+void
+poppler_page_transition_free (PopplerPageTransition *transition)
+{
+  g_free (transition);
+}
+
+/* Form Field Mapping Type */
+GType
+poppler_form_field_mapping_get_type (void)
+{
+  static GType our_type = 0;
+  if (our_type == 0)
+    our_type = g_boxed_type_register_static("PopplerFormFieldMapping",
+					    (GBoxedCopyFunc) poppler_form_field_mapping_copy,
+					    (GBoxedFreeFunc) poppler_form_field_mapping_free);
+  return our_type;
+}
+
+PopplerFormFieldMapping *
+poppler_form_field_mapping_new (void)
+{
+  return (PopplerFormFieldMapping *) g_new0 (PopplerFormFieldMapping, 1);
+}
+
+PopplerFormFieldMapping *
+poppler_form_field_mapping_copy (PopplerFormFieldMapping *mapping)
+{
+  PopplerFormFieldMapping *new_mapping;
+  
+  new_mapping = poppler_form_field_mapping_new ();
+  *new_mapping = *mapping;
+
+  if (mapping->field)
+	  new_mapping->field = (PopplerFormField *)g_object_ref (mapping->field);
+	    
+  return new_mapping;
+}
+
+void
+poppler_form_field_mapping_free (PopplerFormFieldMapping *mapping)
+{
+  if (!mapping)
+    return;
+
+  if (mapping->field)
+    g_object_unref (mapping->field);
+  
+  g_free (mapping);
+}
+
+void 
+poppler_page_get_crop_box (PopplerPage *page, PopplerRectangle *rect)
+{
+  PDFRectangle* cropBox = page->page->getCropBox ();
+  
+  rect->x1 = cropBox->x1;
+  rect->x2 = cropBox->x2;
+  rect->y1 = cropBox->y1;
+  rect->y2 = cropBox->y2;
+}
+
