@@ -52,9 +52,24 @@ static void str_term_source(j_decompress_ptr cinfo)
 {
 }
 
-DCTStream::DCTStream(Stream *strA):
+DCTStream::DCTStream(Stream *strA, int colorXformA) :
   FilterStream(strA) {
-  
+  init();
+}
+
+DCTStream::~DCTStream() {
+  jpeg_destroy_decompress(&cinfo);
+  delete str;
+}
+
+void exitErrorHandler(jpeg_common_struct *error) {
+  j_decompress_ptr cinfo = (j_decompress_ptr)error;
+  str_src_mgr * src = (struct str_src_mgr *)cinfo->src;
+  src->abort = true;
+}
+
+void DCTStream::init()
+{
   jpeg_create_decompress(&cinfo);
   src.pub.init_source = str_init_source;
   src.pub.fill_input_buffer = str_fill_input_buffer;
@@ -65,21 +80,26 @@ DCTStream::DCTStream(Stream *strA):
   src.pub.next_input_byte = NULL;
   src.str = str;
   src.index = 0;
+  src.abort = false;
   cinfo.src = (jpeg_source_mgr *)&src;
-  cinfo.err = jpeg_std_error(&jerr);
+  jpeg_std_error(&jerr);
+  jerr.error_exit = &exitErrorHandler;
+  cinfo.err = &jerr;
   x = 0;
-}
-
-DCTStream::~DCTStream() {
-  jpeg_destroy_decompress(&cinfo);
-  delete str;
+  row_buffer = NULL;
 }
 
 void DCTStream::reset() {
   int row_stride;
 
   str->reset();
-  
+
+  if (row_buffer)
+  {
+    jpeg_destroy_decompress(&cinfo);
+    init();
+  }
+
   // JPEG data has to start with 0xFF 0xD8
   // but some pdf like the one on 
   // https://bugs.freedesktop.org/show_bug.cgi?id=3299
@@ -112,6 +132,8 @@ void DCTStream::reset() {
   }
 
   jpeg_read_header(&cinfo, TRUE);
+  if (src.abort) return;
+
   jpeg_start_decompress(&cinfo);
 
   row_stride = cinfo.output_width * cinfo.output_components;
@@ -119,6 +141,8 @@ void DCTStream::reset() {
 }
 
 int DCTStream::getChar() {
+  if (src.abort) return EOF;
+  
   int c;
 
   if (x == 0) {
@@ -136,6 +160,8 @@ int DCTStream::getChar() {
 }
 
 int DCTStream::lookChar() {
+  if (src.abort) return EOF;
+  
   int c;
   c = row_buffer[0][x];
   return c;
