@@ -17,21 +17,16 @@
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
-#define UNSTABLE_POPPLER_QT4
-
 #include "poppler-qt4.h"
 
+#include <config.h>
 #include <ErrorCodes.h>
 #include <GlobalParams.h>
 #include <Outline.h>
 #include <PDFDoc.h>
-#include <PSOutputDev.h>
 #include <Stream.h>
-#include <UGooString.h>
 #include <Catalog.h>
 
-#include <splash/SplashBitmap.h>
-
 #include <QtCore/QDebug>
 #include <QtCore/QFile>
 #include <QtGui/QImage>
@@ -49,29 +44,39 @@ namespace Poppler {
 	DocumentData *doc = new DocumentData(new GooString(QFile::encodeName(filePath)), 
 					     new GooString(ownerPassword.data()),
 					     new GooString(userPassword.data()));
+	return checkDocument(doc);
+    }
+
+    Document *Document::loadFromData(const QByteArray &fileContents,
+			      const QByteArray &ownerPassword,
+			      const QByteArray &userPassword)
+    {
+	// create stream
+	DocumentData *doc = new DocumentData(fileContents,
+					     new GooString(ownerPassword.data()),
+					     new GooString(userPassword.data()));
+	return checkDocument(doc);
+    }
+    
+    Document *Document::checkDocument(DocumentData *doc)
+    {
 	Document *pdoc;
-	if (doc->doc.isOk() || doc->doc.getErrorCode() == errEncrypted) {
-	    pdoc = new Document(doc);
-	    if (doc->doc.getErrorCode() == errEncrypted)
-		pdoc->m_doc->locked = true;
-	    else
-		pdoc->m_doc->locked = false;
-	    pdoc->m_doc->m_fontInfoScanner = new FontInfoScanner(&(doc->doc));
-	    int numEmb = doc->doc.getCatalog()->numEmbeddedFiles();
-	    if (!(0 == numEmb)) {
-		// we have some embedded documents, build the list
-		for (int yalv = 0; yalv < numEmb; ++yalv) {
-		    EmbFile *ef = doc->doc.getCatalog()->embeddedFile(yalv);
-		    pdoc->m_doc->m_embeddedFiles.append(new EmbeddedFile(ef));
-		    delete ef;
+	if (doc->doc->isOk() || doc->doc->getErrorCode() == errEncrypted) {
+		pdoc = new Document(doc);
+		if (doc->doc->getErrorCode() == errEncrypted)
+			pdoc->m_doc->locked = true;
+		else
+		{
+			pdoc->m_doc->locked = false;
+			pdoc->m_doc->fillMembers();
 		}
-	    }
-	    return pdoc;
+		return pdoc;
 	}
 	else
-	    delete doc;
-	    return NULL;
-
+	{
+		delete doc;
+	}
+	return NULL;
     }
 
     Document::Document(DocumentData *dataA)
@@ -84,6 +89,11 @@ namespace Poppler {
 	delete m_doc;
     }
 
+    Page *Document::page(int index) const
+    {
+	return new Page(this, index);
+    }
+
     bool Document::isLocked() const
     {
 	return m_doc->locked;
@@ -94,19 +104,26 @@ namespace Poppler {
     {
 	if (m_doc->locked) {
 	    /* racier then it needs to be */
-	    GooString *ownerPwd = new GooString(ownerPassword.data());
-	    GooString *userPwd = new GooString(userPassword.data());
-	    DocumentData *doc2 = new DocumentData(m_doc->doc.getFileName(),
-						  ownerPwd,
-						  userPwd);
-	    delete ownerPwd;
-	    delete userPwd;
-	    if (!doc2->doc.isOk()) {
+	    DocumentData *doc2;
+	    if (!m_doc->fileContents.isEmpty())
+	    {
+		doc2 = new DocumentData(m_doc->fileContents,
+					new GooString(ownerPassword.data()),
+					new GooString(userPassword.data()));
+	    }
+	    else
+	    {
+		doc2 = new DocumentData(new GooString(m_doc->doc->getFileName()),
+					new GooString(ownerPassword.data()),
+					new GooString(userPassword.data()));
+	    }
+	    if (!doc2->doc->isOk()) {
 		delete doc2;
 	    } else {
 		delete m_doc;
 		m_doc = doc2;
 		m_doc->locked = false;
+		m_doc->fillMembers();
 	    }
 	}
 	return m_doc->locked;
@@ -114,7 +131,7 @@ namespace Poppler {
 
     Document::PageMode Document::pageMode(void) const
     {
-	switch (m_doc->doc.getCatalog()->getPageMode()) {
+	switch (m_doc->doc->getCatalog()->getPageMode()) {
 	case Catalog::pageModeNone:
 	    return UseNone;
 	case Catalog::pageModeOutlines:
@@ -134,7 +151,7 @@ namespace Poppler {
 
     Document::PageLayout Document::pageLayout(void) const
     {
-	switch (m_doc->doc.getCatalog()->getPageLayout()) {
+	switch (m_doc->doc->getCatalog()->getPageLayout()) {
 	case Catalog::pageLayoutNone:
 	    return NoLayout;
 	case Catalog::pageLayoutSinglePage:
@@ -156,7 +173,7 @@ namespace Poppler {
 
     int Document::numPages() const
     {
-	return m_doc->doc.getNumPages();
+	return m_doc->doc->getNumPages();
     }
 
     QList<FontInfo> Document::fonts() const
@@ -166,7 +183,7 @@ namespace Poppler {
 	return ourList;
     }
 
-    const QList<EmbeddedFile*> &Document::embeddedFiles() const
+    QList<EmbeddedFile*> Document::embeddedFiles() const
     {
 	return m_doc->m_embeddedFiles;
     }
@@ -194,7 +211,7 @@ namespace Poppler {
 	if ( m_doc->locked )
 	    return NULL;
 
-	m_doc->doc.getDocInfo( &info );
+	m_doc->doc->getDocInfo( &info );
 	if ( !info.isDict() )
 	    return NULL;
 
@@ -250,14 +267,16 @@ namespace Poppler {
 	if ( m_doc->locked )
 	    return QStringList();
 
-	m_doc->doc.getDocInfo( &info );
+	m_doc->doc->getDocInfo( &info );
 	if ( !info.isDict() )
 	    return QStringList();
 
 	Dict *infoDict = info.getDict();
 	// somehow iterate over keys in infoDict
 	for( int i=0; i < infoDict->getLength(); ++i ) {
-	    keys.append( QString::fromAscii(infoDict->getKey(i)->getCString()) );
+	    const char *aux = infoDict->getKey(i);
+	    keys.append( QString::fromAscii(aux) );
+	    delete[] aux;
 	}
 
 	info.free();
@@ -272,7 +291,7 @@ namespace Poppler {
 	    return QDateTime();
 
 	Object info;
-	m_doc->doc.getDocInfo( &info );
+	m_doc->doc->getDocInfo( &info );
 	if ( !info.isDict() ) {
 	    info.free();
 	    return QDateTime();
@@ -284,9 +303,8 @@ namespace Poppler {
 
 	if ( infoDict->lookup( type.toLatin1().data(), &obj )->isString() )
 	{
-	    char *aux = UGooString(*obj.getString()).getCString();
+	    char *aux = obj.getString()->getCString();
 	    result = Poppler::convertDate(aux);
-	    delete[] aux;
 	}
 	obj.free();
 	info.free();
@@ -295,42 +313,42 @@ namespace Poppler {
 
     bool Document::isEncrypted() const
     {
-	return m_doc->doc.isEncrypted();
+	return m_doc->doc->isEncrypted();
     }
 
     bool Document::isLinearized() const
     {
-	return m_doc->doc.isLinearized();
+	return m_doc->doc->isLinearized();
     }
 
     bool Document::okToPrint() const
     {
-	return m_doc->doc.okToPrint();
+	return m_doc->doc->okToPrint();
     }
 
     bool Document::okToPrintHighRes() const
     {
-	return m_doc->doc.okToPrintHighRes();
+	return m_doc->doc->okToPrintHighRes();
     }
 
     bool Document::okToChange() const
     {
-	return m_doc->doc.okToChange();
+	return m_doc->doc->okToChange();
     }
 
     bool Document::okToCopy() const
     {
-	return m_doc->doc.okToCopy();
+	return m_doc->doc->okToCopy();
     }
 
     bool Document::okToAddNotes() const
     {
-	return m_doc->doc.okToAddNotes();
+	return m_doc->doc->okToAddNotes();
     }
 
     bool Document::okToFillForm() const
     {
-	return m_doc->doc.okToFillForm();
+	return m_doc->doc->okToFillForm();
     }
 
     bool Document::okToCreateFormFields() const
@@ -340,25 +358,25 @@ namespace Poppler {
 
     bool Document::okToExtractForAccessibility() const
     {
-	return m_doc->doc.okToAccessibility();
+	return m_doc->doc->okToAccessibility();
     }
 
     bool Document::okToAssemble() const
     {
-	return m_doc->doc.okToAssemble();
+	return m_doc->doc->okToAssemble();
     }
 
     double Document::pdfVersion() const
     {
-	return m_doc->doc.getPDFVersion();
+	return m_doc->doc->getPDFVersion();
     }
 
-    Page *Document::page(QString label) const
+    Page *Document::page(const QString &label) const
     {
 	GooString label_g(label.toAscii().data());
 	int index;
 
-	if (!m_doc->doc.getCatalog()->labelToIndex (&label_g, &index))
+	if (!m_doc->doc->getCatalog()->labelToIndex (&label_g, &index))
 	    return NULL;
 
 	return page(index);
@@ -366,12 +384,12 @@ namespace Poppler {
 
     bool Document::hasEmbeddedFiles() const
     {
-	return (!(0 == m_doc->doc.getCatalog()->numEmbeddedFiles()));
+	return (!(0 == m_doc->doc->getCatalog()->numEmbeddedFiles()));
     }
     
     QDomDocument *Document::toc() const
     {
-        Outline * outline = m_doc->doc.getOutline();
+        Outline * outline = m_doc->doc->getOutline();
         if ( !outline )
             return NULL;
 
@@ -388,34 +406,16 @@ namespace Poppler {
 
     LinkDestination *Document::linkDestination( const QString &name )
     {
-        UGooString * namedDest = QStringToUGooString( name );
+        if ( m_doc->getOutputDev() == NULL )
+            return NULL;
+
+        GooString * namedDest = QStringToGooString( name );
         LinkDestinationData ldd(NULL, namedDest, m_doc);
         LinkDestination *ld = new LinkDestination(ldd);
         delete namedDest;
         return ld;
     }
     
-    bool Document::print(const QString &file, const QList<int> pageList, double hDPI, double vDPI, int rotate, int paperWidth, int paperHeight)
-    {
-        PSOutputDev *psOut = new PSOutputDev(file.toLatin1().data(), m_doc->doc.getXRef(), m_doc->doc.getCatalog(), 1, m_doc->doc.getNumPages(), psModePS, paperWidth, paperHeight, gFalse);
-
-        if (psOut->isOk())
-        {
-            foreach(int page, pageList)
-            {
-                m_doc->doc.displayPage(psOut, page, hDPI, vDPI, rotate, gFalse, globalParams->getPSCrop(), gFalse);
-            }
-
-            delete psOut;
-            return true;
-        }
-        else
-        {
-            delete psOut;
-            return false;
-        }
-    }
-    
     void Document::setPaperColor(const QColor &color)
     {
         m_doc->setPaperColor(color);
@@ -426,6 +426,69 @@ namespace Poppler {
     	return m_doc->paperColor;
     }
 
+    void Document::setRenderBackend( Document::RenderBackend backend )
+    {
+        // no need to delete the outputdev as for the moment we always create a splash one
+        // as the arthur one does not allow "precaching" due to it's signature
+        // delete m_doc->m_outputDev;
+        // m_doc->m_outputDev = NULL;
+        m_doc->m_backend = backend;
+    }
+
+    Document::RenderBackend Document::renderBackend() const
+    {
+        return m_doc->m_backend;
+    }
+
+    QSet<Document::RenderBackend> Document::availableRenderBackends()
+    {
+        QSet<Document::RenderBackend> ret;
+#if defined(HAVE_SPLASH)
+        ret << Document::SplashBackend;
+        ret << Document::ArthurBackend;
+#endif
+        return ret;
+    }
+
+    void Document::setRenderHint( Document::RenderHint hint, bool on )
+    {
+        if ( on )
+            m_doc->m_hints |= hint;
+        else
+            m_doc->m_hints &= ~(int)hint;
+
+        // the only way to set antialiasing for Splash is on creation
+        if ( m_doc->m_backend == Document::SplashBackend &&
+             ( hint & ( Document::Antialiasing || Document::TextAntialiasing ) ) )
+        {
+            delete m_doc->m_outputDev;
+            m_doc->m_outputDev = NULL;
+        }
+    }
+
+    Document::RenderHints Document::renderHints() const
+    {
+        return Document::RenderHints() & m_doc->m_hints;
+    }
+
+    PSConverter *Document::psConverter() const
+    {
+        return new PSConverter(m_doc);
+    }
+
+    QString Document::metadata() const
+    {
+        QString result;
+        Catalog *catalog = m_doc->doc->getCatalog();
+        if (catalog && catalog->isOk())
+        {
+            GooString *s = catalog->readMetadata();
+            if (s) result = UnicodeParsedString(s);
+            delete s;
+        }
+        return result;
+    }
+
     QDateTime convertDate( char *dateString )
     {
         int year;
