@@ -20,22 +20,22 @@
 #ifndef __POPPLER_QT_H__
 #define __POPPLER_QT_H__
 
-#ifdef UNSTABLE_POPPLER_QT4
+#include "poppler-annotation.h"
+#include "poppler-link.h"
+#include "poppler-page-transition.h"
 
 #include <QtCore/QByteArray>
 #include <QtCore/QDateTime>
+#include <QtCore/QSet>
 #include <QtCore/QVector>
 #include <QtGui/QPixmap>
 #include <QtXml/QDomDocument>
 
-#include <poppler-annotation.h>
-#include <poppler-link.h>
-#include <poppler-page-transition.h>
-
 class EmbFile;
+class Sound;
 
 /**
-   The Poppler Qt bindings
+   The %Poppler Qt4 binding.
 */
 namespace Poppler {
 
@@ -43,10 +43,14 @@ namespace Poppler {
 
     class PageData;
 
+    class FormField;
+
     class TextBoxData;
 
+    class PSConverter;
+
     /**
-        describes the physical location of text on a document page
+        Describes the physical location of text on a document page
        
         This very simple class describes the physical location of text
 	on the page. It consists of
@@ -58,22 +62,26 @@ namespace Poppler {
     friend class Page;
     public:
       /**
-	 The default constructor sets the text and the rectangle the
-	 contains the text. Coordinated for the bBox are in points =
+	 The default constructor sets the \p text and the rectangle that
+	 contains the text. Coordinated for the \p bBox are in points =
 	 1/72 of an inch.
       */
       TextBox(const QString& text, const QRectF &bBox);
+      /**
+	Destructor.
+      */
+      ~TextBox();
 
       /**
-	  returns the text
-      */      
-      const QString &text() const;
+	  Returns the text of this text box
+      */
+      QString text() const;
 
       /**
-	  returns the position of the text, in point, i.e., 1/72 of
+	  Returns the position of the text, in point, i.e., 1/72 of
 	 an inch
       */
-      const QRectF &boundingBox() const;
+      QRectF boundingBox() const;
 
       TextBox *nextWord() const;
 
@@ -82,46 +90,59 @@ namespace Poppler {
       bool hasSpaceAfter() const;
 
     private:
+        Q_DISABLE_COPY(TextBox)
+
 	TextBoxData *m_data;
     };
 
 
+    class FontInfoData;
     /**
        Container class for information about a font within a PDF
        document
     */
-    class FontInfoData;
     class FontInfo {
     public:
 	enum Type {
-	    unknown = 0,
-	    Type1 = 1,
-	    Type1C = 2,
-	    Type3 = 3,
-	    TrueType = 4,
-	    CIDType0 = 5,
-	    CIDType0C = 6,
-	    CIDTrueType = 7
+		unknown,
+		Type1,
+		Type1C,
+		Type1COT,
+		Type3,
+		TrueType,
+		TrueTypeOT,
+		CIDType0,
+		CIDType0C,
+		CIDType0COT,
+		CIDTrueType,
+		CIDTrueTypeOT
 	};
 	
 	/**
-	   Create a new font information container
+	   Create a new font information container.
 	*/
 	FontInfo( const FontInfoData &fid );
 	
+	/**
+	   Copy constructor.
+	*/
 	FontInfo( const FontInfo &fi );
 	
+	/**
+	   Destructor.
+	*/
 	~FontInfo();
 
 	/**
 	   The name of the font. Can be QString::null if the font has no name
 	*/
-	const QString &name() const;
+	QString name() const;
 
 	/**
-	   The path of the font file used to represent this font on this system.
+	   The path of the font file used to represent this font on this system,
+	   or a null string is the font is embedded
 	*/
-	const QString &file() const;
+	QString file() const;
 
 	/**
 	   Whether the font is embedded in the file, or not
@@ -155,26 +176,28 @@ namespace Poppler {
 
 	   \sa type for a enumeration version
 	*/
-	const QString typeName() const;
+	QString typeName() const;
+
+	FontInfo& operator=( const FontInfo &fi );
 
     private:
 	FontInfoData *m_data;
     };
 
 
+    class EmbeddedFileData;
     /**
        Container class for an embedded file with a PDF document
     */
-    class EmbeddedFileData;
     class EmbeddedFile {
     public:
-	/**
-	   Create a new embedded file container
-	*/
+	/// \cond PRIVATE
 	EmbeddedFile(EmbFile *embfile);
+	/// \endcond
 	
-	EmbeddedFile(const EmbeddedFile &ef);
-	
+	/**
+	   Destructor.
+	*/
 	~EmbeddedFile();
 
 	/**
@@ -189,6 +212,13 @@ namespace Poppler {
 	*/
 	QString description() const;
 
+	/**
+	   The size of the file.
+	
+	   This will return < 0 if there is no size element
+	*/
+	int size() const;
+
 	/**
 	   The modification date for the embedded file, if known.
 	*/
@@ -198,9 +228,16 @@ namespace Poppler {
 	   The creation date for the embedded file, if known.
 	*/
 	QDateTime createDate() const;
+	
+	/**
+		The checksum of the file.
+		
+		This will return an empty QString if there is no checksum element
+	*/
+	QByteArray checksum() const;
 
 	/**
-	   The data as an array
+	   The data as a byte array
 	*/
 	QByteArray data();
 
@@ -210,6 +247,8 @@ namespace Poppler {
 	//QDataStream dataStream() const;
 
     private:
+	Q_DISABLE_COPY(EmbeddedFile)
+
 	EmbeddedFileData *m_embeddedFile;
     };
 
@@ -220,21 +259,31 @@ namespace Poppler {
     class Page {
 	friend class Document;
     public:
+	/**
+	   Destructor.
+	*/
 	~Page();
 
+	enum Rotation { Rotate0 = 0, Rotate90 = 1, Rotate180 = 2, Rotate270 = 3 };
+
+	/**
+	   The kinds of page actions
+	*/
+	enum PageAction {
+	    Opening,   ///< The action when a page is "opened"
+	    Closing    ///< The action when a page is "closed"
+	};
+
 	/** 
-	   Render the page to a QImage using the Splash renderer
+	   Render the page to a QImage using the current Document renderer
+	   (see Document::renderBackend())
 	   
-	   This method can be used to render the page to a QPixmap. It
-	   uses the "Splash" rendering engine. This method is reasonably
-	   well-tested and has produced good output so far.
-	   
-	   If x=y=w=h=-1, the method will automatically compute the
-           size of the pixmap from the horizontal and vertical
-           resolutions specified in xres and yres. Otherwise, the
+	   If \p x = \p y = \p w = \p h = -1, the method will automatically
+           compute the size of the image from the horizontal and vertical
+           resolutions specified in \p xres and \p yres. Otherwise, the
            method renders only a part of the page, specified by the
-           parameters (x, y, w, h) in pixel coordinates. The QPixmap
-           returned then has size (w, h), independent of the page
+           parameters (\p x, \p y, \p w, \p h) in pixel coordinates. The returned
+           QImage then has size (\p w, \p h), independent of the page
            size.
 
 	   \param x specifies the left x-coordinate of the box, in
@@ -253,62 +302,18 @@ namespace Poppler {
 	   \param yres vertical resolution of the graphics device, in
 	   dots per inch
 	
-	   \param doLinks calculate links
+	   \param rotate how to rotate the page
 
-	   \warning The parameter (x,y,w,h) are not
+	   \note if the current Document renderer does not appear among the
+	   Document::availableRenderBackends(), the result is \em always a null QImage.
+
+	   \warning The parameter (\p x, \p y, \p w, \p h) are not
 	   well-tested. Unusual or meaningless parameters may lead to
 	   rather unexpected results.
 
 	   \returns a QImage of the page, or a null image on failure.
         */
-	QImage splashRenderToImage(double xres=72.0, double yres=72.0, int x=-1, int y=-1, int w=-1, int h=-1, bool doLinks = false) const;
-	
-	/**
-	   Render the page to a QPixmap using the Splash renderer
-	   
-	   This member function is provided for convenience. It behaves essentially like the above function.
-	   It is used as follows.
-
-@code
-Poppler::Page* pdfPage;
-
-// Generate a QPixmap of the rendered page
-QPixmap* pixmap = pdfPage->splashRenderToPixmap(0, 0, 0, 0, xres, yres );
-if (pixmap == 0) {
-  ... error message ...
-  return;
-}
-
-... use pixmap ...
-
-delete pixmap;
-@endcode
-	 */
-	QPixmap *splashRenderToPixmap(double xres=72.0, double yres=72.0, int x=-1, int y=-1, int w=-1, int h=-1, bool doLinks = false) const;
-
-	/**
-	    Render the page to a pixmap using the Arthur (Qt4) renderer
-	   
-	   \param q pointer to a QPixmap that is already set to the
-	   intended size.
-
-	   \param xres horizontal resolution of the graphics device,
-	   in dots per inch
-
-	   \param yres vertical resolution of the graphics device, in
-	   dots per inch
-	   
-	   You are meant to create the pixmap before passing it to
-	   this routine, using something like:
-	   \code
-	   QPixmap* myPixmap = new QPixmap(page->pageSize());
-	   page->renderToPixmap(myPixmap);
-	   \endcode
-
-	   \warning This is a work in progress. Results are unlikely to be of
-	   high quality. 
-	*/
-        void renderToPixmap(QPixmap *q, double xres=72.0, double yres=72.0) const;
+	QImage renderToImage(double xres=72.0, double yres=72.0, int x=-1, int y=-1, int w=-1, int h=-1, Rotation rotate = Rotate0) const;
 
 	/**
 	   Returns the text that is inside a specified rectangle
@@ -332,7 +337,7 @@ delete pixmap;
 	   \param direction in which direction do the search
 	   \param caseSensitive be case sensitive?
 	**/
-	bool search(const QString &text, QRectF &rect, SearchDirection direction, SearchMode caseSensitive) const;
+	bool search(const QString &text, QRectF &rect, SearchDirection direction, SearchMode caseSensitive, Rotation rotate = Rotate0) const;
 	
 
 	/**
@@ -348,7 +353,7 @@ delete pixmap;
 
 	   \warning This method is not tested with Asian scripts
 	*/
-	QList<TextBox*> textList() const;
+	QList<TextBox*> textList(Rotation rotate = Rotate0) const;
 
 	/**
 	   \return The dimensions of the page, in points (i.e. 1/72th on an inch)
@@ -356,14 +361,14 @@ delete pixmap;
 	QSizeF pageSizeF() const;
 
 	/**
-	   The dimensions of the page, in pixels.
+	   The size of the page, in pixels
 	*/
 	QSize pageSize() const;
 
 	/**
-	  \brief Returns the transition of this page
+	  Returns the transition of this page
 
-	  @returns a pointer to a PageTransition structure that
+	  \returns a pointer to a PageTransition structure that
 	  defines how transition to this page shall be performed. The
 	  PageTransition structure is owned by this page, and will
 	  automatically be destroyed when this page class is
@@ -371,6 +376,11 @@ delete pixmap;
 	**/
 	PageTransition *transition() const;
 	
+	/**
+	  Gets the page action specified, or NULL if there is no action
+	**/
+	Link *action( PageAction act ) const;
+	
 	/**
 	   Types of orientations that are possible
 	*/
@@ -392,16 +402,35 @@ delete pixmap;
 	void defaultCTM(double *CTM, double dpiX, double dpiY, int rotate, bool upsideDown);
 	
 	/**
-	  Gets the links of the page once it has been rendered if doLinks was true
+	  Gets the links of the page
 	*/
 	QList<Link*> links() const;
 	
 	/**
 	 Returns the annotations of the page
 	*/
-	QList<Annotation*> annotations () const;
+	QList<Annotation*> annotations() const;
+
+	/**
+	 Returns the form fields on the page
+	*/
+	QList<FormField*> formFields() const;
+
+	/**
+	 Returns the page duration. That is the time, in seconds, that the page
+	 should be displayed before the presentation automatically advances to the next page.
+	 Returns < 0 if duration is not set.
+	*/
+	double duration() const;
+	
+	/**
+	   Returns the label of the page, or a null string is the page has no label.
+	**/
+	QString label() const;
 	
     private:
+	Q_DISABLE_COPY(Page)
+
 	Page(const Document *doc, int index);
 	PageData *m_page;
     };
@@ -439,34 +468,70 @@ delete pixmap;
 	    TwoColumnLeft, ///< Display the pages in two columns, with odd-numbered pages on the left
 	    TwoColumnRight, ///< Display the pages in two columns, with odd-numbered pages on the right
 	    TwoPageLeft, ///< Display the pages two at a time, with odd-numbered pages on the left
-	    TwoPageRight, ///< Display the pages two at a time, with odd-numbered pages on the right
+	    TwoPageRight ///< Display the pages two at a time, with odd-numbered pages on the right
 	};
 
+	/**
+	   The render backends available
+	*/
+	enum RenderBackend {
+	    SplashBackend,   ///< Splash backend
+	    ArthurBackend   ///< Arthur (Qt4) backend
+	};
+
+	/**
+	   The render hints available
+	*/
+	enum RenderHint {
+	    Antialiasing = 0x00000001,      ///< Antialiasing for graphics
+	    TextAntialiasing = 0x00000002   ///< Antialiasing for text
+	};
+	Q_DECLARE_FLAGS( RenderHints, RenderHint )
+
 	/**
 	   Load the document from a file on disk
 
 	   \param filePath the name (and path, if required) of the file to load
 
+	   \return NULL on error
+
 	   \warning The application owns the pointer to Document, and this should
 	   be deleted when no longer required.
+	
+	   \warning The returning document may be locked.
 	*/
 	static Document *load(const QString & filePath,
 			      const QByteArray &ownerPassword=QByteArray(),
 			      const QByteArray &userPassword=QByteArray());
+	
+	/**
+	   Load the document from memory
+
+	   \param fileContents the file contents. They are copied so there is no need 
+	                       to keep the byte array around for the full life time of 
+	                       the document.
+
+	   \warning The application owns the pointer to Document, and this should
+	   be deleted when no longer required.
+	*/
+	static Document *loadFromData(const QByteArray &fileContents,
+			      const QByteArray &ownerPassword=QByteArray(),
+			      const QByteArray &userPassword=QByteArray());
   
 	/**
-	   Get a specified page
+	   Get a specified Page
      
 	   Note that this follows the PDF standard of being zero based - if you
 	   want the first page, then you need an index of zero.
 
 	   \param index the page number index
 	*/
-	Page *page(int index) const{ return new Page(this, index); }
+	Page *page(int index) const;
 
 	/**
 	   \overload
 
+
 	   The intent is that you can pass in a label like "ix" and
 	   get the page with that label (which might be in the table of
 	   contents), or pass in "1" and get the page that the user
@@ -475,7 +540,7 @@ delete pixmap;
 
 	   \param label the page label
 	*/
-	Page *page(QString label) const;
+	Page *page(const QString &label) const;
 
 	/**
 	   The number of pages in the document
@@ -511,11 +576,15 @@ delete pixmap;
 
 	   You would use this method with something like:
 	   \code
-	   QDateTime created = m_doc->date("CreationDate");
-	   QDateTime modded = m_doc->date("ModDate");
+QDateTime created = m_doc->date("CreationDate");
+QDateTime modified = m_doc->date("ModDate");
 	   \endcode
 
-	   \param data the type of date that is required (such as CreationDate or ModDate)
+	   The available dates are:
+	   - CreationDate: the date of creation of the document
+	   - ModDate: the date of the last change in the document
+
+	   \param data the type of date that is required
 
 	*/
 	QDateTime date( const QString & data ) const;
@@ -525,12 +594,12 @@ delete pixmap;
 
 	   You would use this method with something like:
 	   \code
-	   QString title = m_doc->info("Title");
-	   QString subject = m_doc->info("Subject");
+QString title = m_doc->info("Title");
+QString subject = m_doc->info("Subject");
 	   \endcode
 
-	   In addition to Title and Subject, other information that may be available
-	   includes Author, Keywords, Creator and Producer.
+	   In addition to \c Title and \c Subject, other information that may
+	   be available include \c Author, \c Keywords, \c Creator and \c Producer.
 
 	   \param data the information that is required
 
@@ -616,8 +685,9 @@ delete pixmap;
 	/**
 	   Test if the permissions on the document allow it to be
 	   "assembled" - insertion, rotation and deletion of pages;
-	   or creation of bookmarks and thumbnail images. This can
-	   be true even if okToChange is false.
+	   or creation of bookmarks and thumbnail images.
+
+	   \note this can be true even if okToChange() is false
 	*/
 	bool okToAssemble() const;
 
@@ -639,6 +709,7 @@ delete pixmap;
 	/**
 	   \overload
 
+
 	   \param numPages the number of pages to scan
 	   \param fontList pointer to the list where the font information
 	   should be placed
@@ -654,7 +725,7 @@ delete pixmap;
 	   \note there are two types of embedded document - this call
 	   only accesses documents that are embedded at the document level.
 	*/
-	const QList<EmbeddedFile*> &embeddedFiles() const;
+	QList<EmbeddedFile*> embeddedFiles() const;
 
 	/**
 	   Whether there are any documents embedded in this PDF document.
@@ -665,12 +736,13 @@ delete pixmap;
 	  Gets the TOC of the Document, it is application responsabiliy to delete
 	  it when no longer needed
 	
-	  * In the tree the tag name is the 'screen' name of the entry. A tag can have
-	  * attributes. Here follows the list of tag attributes with meaning:
-	  * - Destination: A string description of the referred destination
-	  * - DestinationName: A 'named reference' to the viewport that must be converted
-	  *      using linkDestination( *destination_name* )
-	  * - ExternalFileName: A link to a external filename
+	  In the tree the tag name is the 'screen' name of the entry. A tag can have
+	  attributes. Here follows the list of tag attributes with meaning:
+	  - Destination: A string description of the referred destination
+	  - DestinationName: A 'named reference' to the viewport that must be converted
+	       using \p linkDestination( \em destination_name )
+	  - ExternalFileName: A link to a external filename
+	  - Open: A bool value that tells whether the subbranch of the item is open or not
 	
 	  Returns NULL if the Document does not have TOC
 	*/
@@ -678,19 +750,184 @@ delete pixmap;
 	
 	LinkDestination *linkDestination( const QString &name );
 	
-	bool print(const QString &fileName, const QList<int> pageList, double hDPI, double vDPI, int rotate);
-	
 	/**
-	  Sets the paper color, if not set it is white by default.
+	  Sets the paper color
+
+	  \param color the new paper color
 	 */
 	void setPaperColor(const QColor &color);
+	/**
+	  The paper color
+
+	  The default color is white.
+	 */
 	QColor paperColor() const;
 
+	/**
+	 Sets the backend used to render the pages.
+
+	 \note setting a rendering backend that does not appear in the
+	 availableRenderBackends() will always result in null QImage's.
+
+	 \param backend the new rendering backend
+	 */
+	void setRenderBackend( RenderBackend backend );
+	/**
+	  The currently set render backend
+
+	  The default backend is \ref SplashBackend
+	 */
+	RenderBackend renderBackend() const;
+
+	/**
+	  The available rendering backends.
+	 */
+	static QSet<RenderBackend> availableRenderBackends();
+
+	/**
+	 Sets the render \p hint .
+
+	 \note some hints may not be supported by some rendering backends.
+
+	 \param on whether the flag should be added or removed.
+	 */
+	void setRenderHint( RenderHint hint, bool on = true );
+	/**
+	  The currently set render hints.
+	 */
+	RenderHints renderHints() const;
+	
+	/**
+	  Gets a new PS converter for this document.
+
+	  The caller gets the ownership of the returned converter.
+	 */
+	PSConverter *psConverter() const;
+	
+	/**
+	  Gets the metadata stream contents
+	*/
+	QString metadata() const;
+
+	/**
+	   Destructor.
+	*/
 	~Document();
   
     private:
+	Q_DISABLE_COPY(Document)
+
 	DocumentData *m_doc;
+	
 	Document(DocumentData *dataA);
+	static Document *checkDocument(DocumentData *doc);
+    };
+    
+    class PSConverterData;
+    /**
+       Converts a PDF to PS
+
+       Sizes have to be in Points (1/72 inch)
+
+       If you are using QPrinter you can get paper size by doing:
+       \code
+QPrinter dummy(QPrinter::PrinterResolution);
+dummy.setFullPage(true);
+dummy.setPageSize(myPageSize);
+width = dummy.width();
+height = dummy.height();
+       \endcode
+    */
+    class PSConverter
+    {
+        friend class Document;
+        public:
+            /**
+              Destructor.
+            */
+            ~PSConverter();
+
+            /** Sets the output file name. Mandatory. */
+            void setOutputFileName(const QString &outputFileName);
+
+            /** Sets the list of pages to print. Mandatory. */
+            void setPageList(const QList<int> &pageList);
+
+            /**
+              Sets the title of the PS Document. Optional
+            */
+            void setTitle(const QString &title);
+
+            /**
+              Sets the horizontal DPI. Defaults to 72.0
+            */
+            void setHDPI(double hDPI);
+
+            /**
+              Sets the vertical DPI. Defaults to 72.0
+            */
+            void setVDPI(double vDPI);
+
+            /**
+              Sets the rotate. Defaults to not rotated
+            */
+            void setRotate(int rotate);
+
+            /**
+              Sets the output paper width. Has to be set.
+            */
+            void setPaperWidth(int paperWidth);
+
+            /**
+              Sets the output paper height. Has to be set.
+            */
+            void setPaperHeight(int paperHeight);
+
+            /**
+              Sets the output right margin. Defaults to 0
+            */
+            void setRightMargin(int marginRight);
+
+            /**
+              Sets the output bottom margin. Defaults to 0
+            */
+            void setBottomMargin(int marginBottom);
+
+            /**
+              Sets the output left margin. Defaults to 0
+            */
+            void setLeftMargin(int marginLeft);
+
+            /**
+              Sets the output top margin. Defaults to 0
+            */
+            void setTopMargin(int marginTop);
+
+            /**
+              Defines if margins have to be strictly followed (even if that
+              means changing aspect ratio), or if the margins can be adapted
+              to keep aspect ratio.
+
+              Defaults to false.
+            */
+            void setStrictMargins(bool strictMargins);
+
+            /** Defines if the page will be rasterized to an image before printing. Defaults to false */
+            void setForceRasterize(bool forceRasterize);
+
+            /**
+              Does the conversion.
+
+              \return whether the conversion succeeded
+            */
+            bool convert();
+
+        private:
+            Q_DISABLE_COPY(PSConverter)
+
+            PSConverter(DocumentData *document);
+
+            PSConverterData *m_data;
     };
 
     /**
@@ -698,7 +935,81 @@ delete pixmap;
     */
     QDateTime convertDate( char *dateString );
 
+    class SoundData;
+    /**
+       Container class for a sound file in a PDF document.
+
+	A sound can be either External (in that case should be loaded the file
+       whose url is represented by url() ), or Embedded, and the player has to
+       play the data contained in data().
+    */
+    class SoundObject {
+    public:
+	/**
+	   The type of sound
+	*/
+	enum SoundType {
+	    External,     ///< The real sound file is external
+	    Embedded      ///< The sound is contained in the data
+	};
+
+	/**
+	   The encoding format used for the sound
+	*/
+	enum SoundEncoding {
+	    Raw,          ///< Raw encoding, with unspecified or unsigned values in the range [ 0, 2^B - 1 ]
+	    Signed,       ///< Twos-complement values
+	    muLaw,        ///< mu-law-encoded samples
+	    ALaw          ///< A-law-encoded samples
+	};
+
+	/// \cond PRIVATE
+	SoundObject(Sound *popplersound);
+	/// \endcond
+	
+	~SoundObject();
+
+	/**
+	   Is the sound embedded (\ref Embedded ) or external (\ref External )?
+	*/
+	SoundType soundType() const;
+
+	/**
+	   The URL of the sound file to be played, in case of \ref External
+	*/
+	QString url() const;
+
+	/**
+	   The data of the sound, in case of \ref Embedded
+	*/
+	QByteArray data() const;
+
+	/**
+	   The sampling rate of the sound
+	*/
+	double samplingRate() const;
+
+	/**
+	   The number of sound channels to use to play the sound
+	*/
+	int channels() const;
+
+	/**
+	   The number of bits per sample value per channel
+	*/
+	int bitsPerSample() const;
+
+	/**
+	   The encoding used for the sound
+	*/
+	SoundEncoding soundEncoding() const;
+
+    private:
+	Q_DISABLE_COPY(SoundObject)
+
+	SoundData *m_soundData;
+    };
+
 }
-#endif
 
 #endif
