@@ -16,19 +16,11 @@
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
-#define UNSTABLE_POPPLER_QT4
-
 #include "poppler-qt4.h"
+#include "poppler-private.h"
 
 namespace Poppler {
 
-class TextBoxData
-{
-	public:
-		QString text;
-		QRectF bBox;
-};
-
 TextBox::TextBox(const QString& text, const QRectF &bBox)
 {
 	m_data = new TextBoxData();
@@ -36,14 +28,34 @@ TextBox::TextBox(const QString& text, const QRectF &bBox)
 	m_data->bBox = bBox;
 }
 
-const QString &TextBox::text() const
+TextBox::~TextBox()
+{
+	delete m_data;
+}
+
+QString TextBox::text() const
 {
 	return m_data->text;
 }
 
-const QRectF &TextBox::boundingBox() const
+QRectF TextBox::boundingBox() const
 {
 	return m_data->bBox;
 };
 
+TextBox *TextBox::nextWord() const
+{
+	return m_data->nextWord;
+}
+
+double TextBox::edge(int i) const
+{
+	return m_data->edge[i];
+}
+
+bool TextBox::hasSpaceAfter() const
+{
+	return m_data->hasSpaceAfter;
+}
+
 }
