@@ -3,7 +3,6 @@
 #include <QtGui/QtGui>
 #include <ctype.h>
 
-#define UNSTABLE_POPPLER_QT4
 #include <poppler-qt4.h>
 
 class PDFDisplay : public QWidget           // picture display widget
@@ -17,16 +16,25 @@ protected:
 private:
     void display();
     int m_currentPage;
-    QPixmap *pixmap;
+    QImage image;
     Poppler::Document *doc;
-    bool useArthur;
+    QString backendString;
 };
 
 PDFDisplay::PDFDisplay( Poppler::Document *d, bool arthur )
 {
     doc = d;
     m_currentPage = 0;
-    useArthur = arthur;
+    if (arthur)
+    {
+        backendString = "Arthur";
+        doc->setRenderBackend(Poppler::Document::ArthurBackend);
+    }
+    else
+    {
+        backendString = "Splash";
+        doc->setRenderBackend(Poppler::Document::SplashBackend);
+    }
     display();
 }
 
@@ -35,18 +43,8 @@ void PDFDisplay::display()
     if (doc) {
         Poppler::Page *page = doc->page(m_currentPage);
         if (page) {
-            pixmap = new QPixmap(page->pageSize());
-            if (useArthur)
-            {
-                qDebug() << "Displaying page using Arthur backend: " << m_currentPage;
-                page->renderToPixmap(pixmap, 72.0, 72.0);
-            }
-            else
-            {
-                qDebug() << "Displaying page using Splash backend: " << m_currentPage;
-		delete pixmap;
-                pixmap = page->splashRenderToPixmap();
-            }
+            qDebug() << "Displaying page using" << backendString << "backend: " << m_currentPage;
+            image = page->renderToImage();
             update();
             delete page;
         }
@@ -58,16 +56,15 @@ void PDFDisplay::display()
 PDFDisplay::~PDFDisplay()
 {
     delete doc;
-    delete pixmap;
 }
 
 void PDFDisplay::paintEvent( QPaintEvent *e )
 {
     QPainter paint( this );                     // paint widget
-    if (pixmap) {
-	paint.drawPixmap(0, 0, *pixmap);
+    if (!image.isNull()) {
+	paint.drawImage(0, 0, image);
     } else {
-	qWarning() << "no pixmap";
+	qWarning() << "null image";
     }
 }
 
@@ -133,6 +130,7 @@ int main( int argc, char **argv )
     qDebug() << "   OK to change: " << doc->okToChange();
     qDebug() << "OK to add notes: " << doc->okToAddNotes();
     qDebug() << "      Page mode: " << doc->pageMode();
+    qDebug() << "       Metadata: " << doc->metadata();
     QStringList fontNameList;
     foreach( Poppler::FontInfo font, doc->fonts() )
 	fontNameList += font.name();
@@ -140,6 +138,7 @@ int main( int argc, char **argv )
 
     Poppler::Page *page = doc->page(0);
     qDebug() << "    Page 1 size: " << page->pageSize().width()/72 << "inches x " << page->pageSize().height()/72 << "inches";
+    delete page;
 
     if (argc == 2 || (argc == 3 && strcmp(argv[2], "-arthur") == 0))
     {
